"""Module for spaCy v2/v3 SpaczzRuler compatibility."""

try:
    from ._spaczzruler import make_spaczz_ruler, SpaczzRuler  # type: ignore
except ImportError:  # pragma: no cover
    try:
        from ._spaczzruler_legacy import make_spaczz_ruler, SpaczzRuler  # type: ignore
    except ImportError:  # pragma: no cover
        raise ImportError(
            (
                "Unable to import spaCy v3 or v2 compatible SpaczzRuler.",
                "Ensure you have spaCy >= 2.2 and/or",
                "please raise an issue at https://github.com/gandersen101/spaczz",
            )
        )


make_spaczz_ruler = make_spaczz_ruler
SpaczzRuler = SpaczzRuler



"""Module for FuzzyMatcher with an API semi-analogous to spaCy's PhraseMatcher."""
from __future__ import annotations

from typing import Any

from spacy.vocab import Vocab

from . import _PhraseMatcher
from ..search.fuzzysearcher import FuzzySearcher


class FuzzyMatcher(_PhraseMatcher):
    """spaCy-like matcher for finding fuzzy matches in Doc objects.

    Fuzzy matches added patterns against the `Doc` object it is called on.
    Accepts labeled patterns in the form of `Doc` objects.

    Attributes:
        defaults: Keyword arguments to be used as default matching settings.
            See `FuzzySearcher` documentation for details.
        name: Class attribute - the name of the matcher.
        type: The kind of matcher object.
        _callbacks:
            On match functions to modify `Doc` objects passed to the matcher.
            Can make use of the matches identified.
        _patterns:
            Patterns added to the matcher. Contains patterns
            and kwargs that should be passed to matching function
            for each labels added.
    """

    name = "fuzzy_matcher"

    def __init__(self: FuzzyMatcher, vocab: Vocab, **defaults: Any) -> None:
        """Initializes the fuzzy matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are currently pure
                Python and do not share vocabulary
                with spacy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings.
                These arguments will become the new defaults for matching.
                See `FuzzySearcher` documentation for details.
        """
        super().__init__(vocab=vocab, **defaults)
        self.type = "fuzzy"
        self._searcher = FuzzySearcher(vocab=vocab)



"""Module for search components."""
from ._phrasesearcher import _PhraseSearcher
from .fuzzysearcher import FuzzySearcher
from .regexsearcher import RegexSearcher
from .similaritysearcher import SimilaritySearcher
from .tokensearcher import TokenSearcher

__all__ = [
    "_PhraseSearcher",
    "FuzzySearcher",
    "RegexSearcher",
    "SimilaritySearcher",
    "TokenSearcher",
]



"""Module for spaCy v3 compatible SpaczzRuler."""
from __future__ import annotations

from collections import defaultdict
from itertools import chain
from logging import exception
from pathlib import Path
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
import warnings

try:
    from spacy.language import Language
    from spacy.pipeline import Pipe
    from spacy.scorer import get_ner_prf
    from spacy.tokens import Doc, Span
    from spacy.training import Example, validate_examples
    from spacy.util import SimpleFrozenDict, SimpleFrozenList
except ImportError:  # pragma: no cover
    raise ImportError(
        (
            "Trying to import spaCy v3 compatible SpaczzRuler from spaCy v2.",
            "Please upgrade or use the SpaczzRuler in _spaczzruler-legacy",
        )
    )
import srsly

from ..exceptions import PatternTypeWarning
from ..matcher import FuzzyMatcher, RegexMatcher, TokenMatcher
from ..regex import RegexConfig
from ..util import ensure_path, nest_defaultdict, read_from_disk, write_to_disk


DEFAULT_ENT_ID_SEP = "||"
simple_frozen_dict = SimpleFrozenDict()
simple_frozen_list = SimpleFrozenList()


@Language.factory(
    "spaczz_ruler",
    assigns=["doc.ents", "token.ent_type", "token.ent_iob"],
    default_config={
        "overwrite_ents": False,
        "ent_id_sep": DEFAULT_ENT_ID_SEP,
        "fuzzy_defaults": simple_frozen_dict,
        "regex_defaults": simple_frozen_dict,
        "token_defaults": simple_frozen_dict,
    },
    default_score_weights={
        "ents_f": 1.0,
        "ents_p": 0.0,
        "ents_r": 0.0,
        "ents_per_type": None,
    },
)
def make_spaczz_ruler(
    # typing nlp with Language causes issue with Pydantic in spaCy integration
    nlp: Any,
    name: str,
    overwrite_ents: bool,
    ent_id_sep: str,
    fuzzy_defaults: Dict[str, Any],
    regex_defaults: Dict[str, Any],
    token_defaults: Dict[str, Any],
) -> SpaczzRuler:
    """Factory method for creating a `SpaczzRuler`."""
    return SpaczzRuler(
        nlp,
        name,
        overwrite_ents=overwrite_ents,
        ent_id_sep=ent_id_sep,
        fuzzy_defaults=fuzzy_defaults,
        regex_defaults=regex_defaults,
        token_defaults=token_defaults,
    )


class SpaczzRuler(Pipe):
    """The `SpaczzRuler` adds fuzzy and multi-token regex matches to spaCy `Doc.ents`.

    It can be combined with other spaCy NER components like the statistical
    `EntityRecognizer` and/or the `EntityRuler` to boost accuracy.
    After initialization, the component is typically added to the pipeline
    using `nlp.add_pipe`.

    Attributes:
        nlp: The shared nlp object to pass the vocab to the matchers
            (not currently used by spaczz matchers) and process fuzzy patterns.
        fuzzy_patterns:
            Patterns added to the fuzzy matcher.
        regex_patterns:
            Patterns added to the regex matcher.
        token_patterns:
            Patterns added to the token matcher
        fuzzy_matcher: The `FuzzyMatcher` instance
            the spaczz ruler will use for fuzzy phrase matching.
        regex_matcher: The `RegexMatcher` instance
            the spaczz ruler will use for regex phrase matching.
        token_matcher: The `TokenMatcher` instance
            the spaczz ruler will use for token matching.
        defaults: Default matching settings for their respective matchers.
    """

    name = "spaczz_ruler"

    def __init__(
        self: SpaczzRuler,
        nlp: Language,
        name: str = "spaczz_ruler",
        *,
        overwrite_ents: bool = False,
        ent_id_sep: str = DEFAULT_ENT_ID_SEP,
        fuzzy_defaults: Dict[str, Any] = simple_frozen_dict,
        regex_defaults: Dict[str, Any] = simple_frozen_dict,
        token_defaults: Dict[str, Any] = simple_frozen_dict,
        regex_config: Union[str, RegexConfig] = "default",
        patterns: Optional[Iterable[Dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the spaczz ruler.

        If `patterns` is supplied here, it needs to be an iterable of spaczz patterns:
        dictionaries with `"label"`, `"pattern"`, and `"type"` keys.
        If the patterns are fuzzy or regex phrase patterns
        they can include the optional `"kwargs"` keys.

        For example, a fuzzy phrase pattern:
        `{'label': 'ORG', 'pattern': 'Apple',
        'type': 'fuzzy', 'kwargs': {'min_r2': 90}}`

        Or, a token pattern:
        `{'label': 'ORG', 'pattern': [{'TEXT': {'FUZZY': 'Apple'}}], 'type': 'token'}`

        Prior to spaczz v0.5, optional parameters had to be prepended with "spaczz_"
        to prevent potential conflicts with other spaCy components.
        As of spaCy v3 this is no longer an issue so prepending optional parameters
        with "spaczz_" is no longer necessary.


        Args:
            nlp: The shared `Language` object to pass the vocab to the matchers
                and process fuzzy patterns.
            name: Instance name of the current pipeline component. Typically
                passed in automatically from the factory when the component is
                added. Used to disable the current entity ruler while creating
                phrase patterns with the nlp object.
            overwrite_ents: If existing entities are present, e.g. entities
                added by the model, overwrite them by matches if necessary.
                Default is `False`.
            ent_id_sep: Separator used internally for entity IDs.
            fuzzy_defaults: Modified default parameters to use with the `FuzzyMatcher`.
                Default is `None`.
            regex_defaults: Modified default parameters to use with the `RegexMatcher`.
                Default is `None`.
            token_defaults: Modified default parameters to use with the `TokenMatcher`.
                Default is `None`.
            regex_config: Should largely be ignored as an artifact of an old spaczz
                design pattern. Will likely be updated in the future.
                Default is `"default"`.
            patterns: Optional patterns to load in. Default is `None`.
            kwargs: For backwards compatibility with "spaczz_" prepended parameters.

        Raises:
            TypeError: If matcher defaults passed are not dictionaries.
        """
        self.nlp = nlp
        self.name = name
        self.overwrite = kwargs.get("spaczz_overwrite_ents", overwrite_ents)
        self.fuzzy_patterns: DefaultDict[str, DefaultDict[str, Any]] = nest_defaultdict(
            list, 2
        )
        self.regex_patterns: DefaultDict[str, DefaultDict[str, Any]] = nest_defaultdict(
            list, 2
        )
        self.token_patterns: DefaultDict[str, List[List[Dict[str, Any]]]] = defaultdict(
            list
        )
        self.ent_id_sep = kwargs.get("spaczz_ent_id_sep", ent_id_sep)
        self._ent_ids: DefaultDict[Any, Any] = defaultdict(dict)
        self.defaults = {}
        default_names = (
            "fuzzy_defaults",
            "regex_defaults",
            "token_defaults",
        )
        fuzzy_defaults = kwargs.get("spaczz_fuzzy_defaults", fuzzy_defaults)
        regex_defaults = kwargs.get("spaczz_regex_defaults", regex_defaults)
        token_defaults = kwargs.get("spaczz_token_defaults", token_defaults)
        for default, name in zip(
            (fuzzy_defaults, regex_defaults, token_defaults), default_names
        ):
            if isinstance(default, dict):
                self.defaults[name] = default
            else:
                raise TypeError(
                    (
                        "Defaults must be a dictionary of keyword arguments,",
                        f"not {type(default)}.",
                    )
                )
        self.fuzzy_matcher = FuzzyMatcher(nlp.vocab, **self.defaults["fuzzy_defaults"])
        self.regex_matcher = RegexMatcher(
            nlp.vocab, regex_config, **self.defaults["regex_defaults"]
        )
        self.token_matcher = TokenMatcher(nlp.vocab, **self.defaults["token_defaults"])
        patterns = kwargs.get("spaczz_patterns", patterns)
        if patterns is not None:
            self.add_patterns(patterns)

    def __call__(self: SpaczzRuler, doc: Doc) -> Doc:
        """Find matches in document and add them as entities.

        Args:
            doc: The Doc object in the pipeline.

        Returns:
            The Doc with added entities, if available.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> doc = nlp.make_doc("My name is Anderson, Grunt")
            >>> ruler.add_patterns([{"label": "NAME", "pattern": "Grant Andersen",
                "type": "fuzzy", "kwargs": {"fuzzy_func": "token_sort"}}])
            >>> doc = ruler(doc)
            >>> "Anderson, Grunt" in [ent.text for ent in doc.ents]
            True
        """
        error_handler = self.get_error_handler()
        try:
            matches, lookup = self.match(doc)
            self.set_annotations(doc, matches, lookup)
            return doc
        except exception as e:  # type: ignore
            error_handler(self.name, self, [doc], e)

    def __contains__(self: SpaczzRuler, label: str) -> bool:
        """Whether a label is present in the patterns."""
        return (
            label in self.fuzzy_patterns
            or label in self.regex_patterns
            or label in self.token_patterns
        )

    def __len__(self: SpaczzRuler) -> int:
        """The number of all patterns added to the ruler."""
        n_fuzzy_patterns = sum(len(p["patterns"]) for p in self.fuzzy_patterns.values())
        n_regex_patterns = sum(len(p["patterns"]) for p in self.regex_patterns.values())
        n_token_patterns = sum(len(p) for p in self.token_patterns.values())
        return n_fuzzy_patterns + n_regex_patterns + n_token_patterns

    @property
    def ent_ids(self: SpaczzRuler) -> Tuple[Optional[str], ...]:
        """All entity ids present in the match patterns id properties.

        Returns:
            The unique string entity ids as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy", "id": "BEAT"}])
            >>> ruler.ent_ids
            ('BEAT',)
        """
        keys = set(self.fuzzy_patterns.keys())
        keys.update(self.regex_patterns.keys())
        keys.update(self.token_patterns.keys())
        all_ent_ids = set()

        for k in keys:
            if self.ent_id_sep in k:
                _, ent_id = self._split_label(k)
                all_ent_ids.add(ent_id)
        all_ent_ids_tuple = tuple(all_ent_ids)
        return all_ent_ids_tuple

    @property
    def labels(self: SpaczzRuler) -> Tuple[str, ...]:
        """All labels present in the ruler.

        Returns:
            The unique string labels as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> ruler.labels
            ('AUTHOR',)
        """
        keys = set(self.fuzzy_patterns.keys())
        keys.update(self.regex_patterns.keys())
        keys.update(self.token_patterns.keys())
        all_labels = set()
        for k in keys:
            if self.ent_id_sep in k:
                label, _ = self._split_label(k)
                all_labels.add(label)
            else:
                all_labels.add(k)
        return tuple(all_labels)

    @property
    def patterns(self: SpaczzRuler) -> List[Dict[str, Any]]:
        """Get all patterns and kwargs that were added to the ruler.

        Returns:
            The original patterns and kwargs, one dictionary for each combination.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "STREET", "pattern": "street_addresses",
                "type": "regex", "kwargs": {"predef": True}}])
            >>> ruler.patterns == [
                {
                    "label": "STREET",
                    "pattern": "street_addresses",
                    "type": "regex",
                    "kwargs": {"predef": True},
                    },
                    ]
            True
        """
        all_patterns = []
        for label, fuzzy_patterns in self.fuzzy_patterns.items():
            for fuzzy_pattern, fuzzy_kwargs in zip(
                fuzzy_patterns["patterns"], fuzzy_patterns["kwargs"]
            ):
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": fuzzy_pattern.text, "type": "fuzzy"}
                if fuzzy_kwargs:
                    p["kwargs"] = fuzzy_kwargs
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        for label, regex_patterns in self.regex_patterns.items():
            for regex_pattern, regex_kwargs in zip(
                regex_patterns["patterns"], regex_patterns["kwargs"]
            ):
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": regex_pattern, "type": "regex"}
                if regex_kwargs:
                    p["kwargs"] = regex_kwargs
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        for label, token_patterns in self.token_patterns.items():
            for token_pattern in token_patterns:
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": token_pattern, "type": "token"}
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        return all_patterns

    def add_patterns(
        self: SpaczzRuler,
        patterns: Iterable[Dict[str, Any]],
    ) -> None:
        """Add patterns to the ruler.

        A pattern must be a spaczz pattern:
        `{label (str), pattern (str or list), type (str),
        optional kwargs (dict[str, Any]), and optional id (str)}`.

        For example, a fuzzy phrase pattern:
        `{'label': 'ORG', 'pattern': 'Apple',
        'type': 'fuzzy', 'kwargs': {'min_r2': 90}}`

        Or, a token pattern:
        `{'label': 'ORG', 'pattern': [{'TEXT': {'FUZZY': 'Apple'}}], 'type': 'token'}`

        To utilize regex flags, use inline flags.

        Kwarg details to be updated.

        Args:
            patterns: The spaczz patterns to add.

        Raises:
            TypeError: If patterns is not an iterable of dictionaries.
            ValueError: If one or more patterns do not conform
                the spaczz pattern structure.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> "AUTHOR" in ruler.labels
            True
        """
        # disable the nlp components after this one in case
        # they hadn't been initialized / deserialised yet
        try:
            current_index = -1
            for i, (_name, pipe) in enumerate(self.nlp.pipeline):
                if self == pipe:
                    current_index = i
                    break
            subsequent_pipes = [
                pipe for pipe in self.nlp.pipe_names[current_index + 1 :]
            ]
        except ValueError:
            subsequent_pipes = []
        with self.nlp.select_pipes(disable=subsequent_pipes):
            token_patterns = []
            fuzzy_pattern_labels = []
            fuzzy_pattern_texts = []
            fuzzy_pattern_kwargs = []
            fuzzy_pattern_ids = []
            regex_pattern_labels = []
            regex_pattern_texts = []
            regex_pattern_kwargs = []
            regex_pattern_ids = []

            for entry in patterns:
                try:
                    if isinstance(entry, dict):
                        if entry["type"] == "fuzzy":
                            fuzzy_pattern_labels.append(entry["label"])
                            fuzzy_pattern_texts.append(entry["pattern"])
                            fuzzy_pattern_kwargs.append(entry.get("kwargs", {}))
                            fuzzy_pattern_ids.append(entry.get("id"))
                        elif entry["type"] == "regex":
                            regex_pattern_labels.append(entry["label"])
                            regex_pattern_texts.append(entry["pattern"])
                            regex_pattern_kwargs.append(entry.get("kwargs", {}))
                            regex_pattern_ids.append(entry.get("id"))
                        elif entry["type"] == "token":
                            token_patterns.append(entry)
                        else:
                            warnings.warn(
                                f"""Spaczz pattern "type" must be "fuzzy", "regex",
                                or "token", not {entry["type"]}. Skipping this pattern.
                                """,
                                PatternTypeWarning,
                            )
                    else:
                        raise TypeError(
                            ("Patterns must either be an iterable of dicts.")
                        )
                except KeyError:
                    raise ValueError(
                        (
                            "One or more patterns do not conform",
                            "to spaczz pattern structure: ",
                            "{label (str), pattern (str or list), type (str),",
                            "optional kwargs (dict[str, Any]),",
                            "and optional id (str)}.",
                        )
                    )

            fuzzy_patterns = []
            for label, pattern, kwargs, ent_id in zip(
                fuzzy_pattern_labels,
                self.nlp.pipe(fuzzy_pattern_texts),
                fuzzy_pattern_kwargs,
                fuzzy_pattern_ids,
            ):
                fuzzy_pattern = {
                    "label": label,
                    "pattern": pattern,
                    "kwargs": kwargs,
                    "type": "fuzzy",
                }
                if ent_id:
                    fuzzy_pattern["id"] = ent_id
                fuzzy_patterns.append(fuzzy_pattern)

            regex_patterns = []
            for label, pattern, kwargs, ent_id in zip(
                regex_pattern_labels,
                regex_pattern_texts,
                regex_pattern_kwargs,
                regex_pattern_ids,
            ):
                regex_pattern = {
                    "label": label,
                    "pattern": pattern,
                    "kwargs": kwargs,
                    "type": "regex",
                }
                if ent_id:
                    regex_pattern["id"] = ent_id
                regex_patterns.append(regex_pattern)

            self._add_patterns(fuzzy_patterns, regex_patterns, token_patterns)

    def clear(self: SpaczzRuler) -> None:
        """Reset all patterns."""
        self.fuzzy_patterns = nest_defaultdict(list, 2)
        self.regex_patterns = nest_defaultdict(list, 2)
        self.token_patterns = defaultdict(list)
        self._ent_ids = defaultdict(dict)

    def initialize(
        self: SpaczzRuler,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Optional[Language] = None,
        patterns: Optional[Iterable[Dict[str, Any]]] = None,
    ) -> None:
        """Initialize the pipe for training.

        Args:
            get_examples: Function that returns a representative sample
                of gold-standard Example objects.
            nlp: The current nlp object the component is part of.
            patterns: The list of patterns.
        """
        self.clear()
        if patterns:
            self.add_patterns(patterns)

    def match(
        self: SpaczzRuler, doc: Doc
    ) -> Tuple[
        List[Tuple[str, int, int]],
        DefaultDict[str, Dict[Tuple[str, int, int], Any]],
    ]:
        """Used in call to find matches in a doc."""
        fuzzy_matches = []
        lookup: DefaultDict[str, Dict[Tuple[str, int, int], Any]] = defaultdict(dict)
        for fuzzy_match in self.fuzzy_matcher(doc):
            current_ratio = fuzzy_match[3]
            best_ratio = lookup["ratios"].get(fuzzy_match[:3], 0)
            if current_ratio > best_ratio:
                fuzzy_matches.append(fuzzy_match[:3])
                lookup["ratios"][fuzzy_match[:3]] = current_ratio
        regex_matches = []
        for regex_match in self.regex_matcher(doc):
            current_counts = regex_match[3]
            best_counts = lookup["counts"].get(regex_match[:3])
            if not best_counts or sum(current_counts) < sum(best_counts):
                regex_matches.append(regex_match[:3])
                lookup["counts"][regex_match[:3]] = current_counts
        token_matches = []
        for token_match in self.token_matcher(doc):
            token_matches.append(token_match[:3])
            lookup["details"][token_match[:3]] = 1
        matches = fuzzy_matches + regex_matches + token_matches
        unique_matches, lookup = self._filter_overlapping_matches(matches, lookup)
        return unique_matches, lookup

    def score(self: SpaczzRuler, examples: Any, **kwargs: Any) -> Any:
        """Pipeline scoring for spaCy compatibility."""
        validate_examples(examples, "SpaczzRuler.score")
        return get_ner_prf(examples)

    def set_annotations(
        self: SpaczzRuler,
        doc: Doc,
        matches: List[Tuple[str, int, int]],
        lookup: DefaultDict[
            str, Dict[Tuple[str, int, int], Union[int, Tuple[int, int, int]]]
        ],
    ) -> None:
        """Modify the document in place."""
        entities = list(doc.ents)
        new_entities = []
        seen_tokens: Set[int] = set()
        for match_id, start, end in matches:
            if any(t.ent_type for t in doc[start:end]) and not self.overwrite:
                continue
            # check for end - 1 here because boundaries are inclusive
            if start not in seen_tokens and end - 1 not in seen_tokens:
                if match_id in self._ent_ids:
                    label, ent_id = self._ent_ids[match_id]
                    span = Span(doc, start, end, label=label)
                    if ent_id:
                        for token in span:
                            token.ent_id_ = ent_id
                else:
                    span = Span(doc, start, end, label=match_id)
                span = self._update_custom_attrs(span, match_id, lookup)
                new_entities.append(span)
                entities = [
                    e for e in entities if not (e.start < end and e.end > start)
                ]
                seen_tokens.update(range(start, end))
        doc.ents = entities + new_entities

    def from_bytes(
        self: SpaczzRuler,
        patterns_bytes: bytes,
        *,
        exclude: Iterable[str] = simple_frozen_list,
    ) -> SpaczzRuler:
        """Load the spaczz ruler from a bytestring.

        Args:
            patterns_bytes : The bytestring to load.
            exclude: For spaCy consistency.

        Returns:
            The loaded spaczz ruler.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> ruler_bytes = ruler.to_bytes()
            >>> new_ruler = SpaczzRuler(nlp)
            >>> new_ruler = new_ruler.from_bytes(ruler_bytes)
            >>> "AUTHOR" in new_ruler
            True
        """
        cfg = srsly.msgpack_loads(patterns_bytes)
        self.clear()
        if isinstance(cfg, dict):
            self.add_patterns(cfg.get("patterns", cfg))
            self.defaults = cfg.get("defaults", {})
            if self.defaults.get("fuzzy_defaults"):
                self.fuzzy_matcher = FuzzyMatcher(
                    self.nlp.vocab, **self.defaults["fuzzy_defaults"]
                )
            if self.defaults.get("regex_defaults"):
                self.regex_matcher = RegexMatcher(
                    self.nlp.vocab, **self.defaults["regex_defaults"]
                )
            if self.defaults.get("token_defaults"):
                self.token_matcher = TokenMatcher(
                    self.nlp.vocab, **self.defaults["token_defaults"]
                )
            self.overwrite = cfg.get("overwrite", False)
            self.ent_id_sep = cfg.get("ent_id_sep", DEFAULT_ENT_ID_SEP)
        else:
            self.add_patterns(cfg)
        return self

    def to_bytes(
        self: SpaczzRuler, *, exclude: Iterable[str] = simple_frozen_list
    ) -> bytes:
        """Serialize the spaczz ruler patterns to a bytestring.

        Args:
            exclude: For spaCy consistency.

        Returns:
            The serialized patterns.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> ruler_bytes = ruler.to_bytes()
            >>> isinstance(ruler_bytes, bytes)
            True
        """
        serial = {
            "overwrite": self.overwrite,
            "ent_id_sep": self.ent_id_sep,
            "patterns": self.patterns,
            "defaults": self.defaults,
        }
        return srsly.msgpack_dumps(serial)

    def from_disk(
        self: SpaczzRuler,
        path: Union[str, Path],
        *,
        exclude: Iterable[str] = simple_frozen_list,
    ) -> SpaczzRuler:
        """Load the spaczz ruler from a file.

        Expects a file containing newline-delimited JSON (JSONL)
        with one entry per line.

        Args:
            path: The JSONL file to load.
            exclude: For spaCy consistency.

        Returns:
            The loaded spaczz ruler.

        Example:
            >>> import os
            >>> import tempfile
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            >>>     ruler.to_disk(f"{tmpdir}/ruler")
            >>>     new_ruler = SpaczzRuler(nlp)
            >>>     new_ruler = new_ruler.from_disk(f"{tmpdir}/ruler")
            >>> "AUTHOR" in new_ruler
            True
        """
        path = ensure_path(path)
        self.clear()
        depr_patterns_path = path.with_suffix(".jsonl")
        if depr_patterns_path.is_file():
            patterns = srsly.read_jsonl(depr_patterns_path)
            self.add_patterns(patterns)
        else:
            cfg = {}
            deserializers_patterns = {
                "patterns": lambda p: self.add_patterns(
                    srsly.read_jsonl(p.with_suffix(".jsonl"))
                )
            }
            deserializers_cfg = {"cfg": lambda p: cfg.update(srsly.read_json(p))}
            read_from_disk(path, deserializers_cfg, {})
            self.overwrite = cfg.get("overwrite", False)
            self.defaults = cfg.get("defaults", {})
            if self.defaults.get("fuzzy_defaults"):
                self.fuzzy_matcher = FuzzyMatcher(
                    self.nlp.vocab, **self.defaults["fuzzy_defaults"]
                )
            if self.defaults.get("regex_defaults"):
                self.regex_matcher = RegexMatcher(
                    self.nlp.vocab, **self.defaults["regex_defaults"]
                )
            if self.defaults.get("token_defaults"):
                self.token_matcher = TokenMatcher(
                    self.nlp.vocab, **self.defaults["token_defaults"]
                )
            self.ent_id_sep = cfg.get("ent_id_sep", DEFAULT_ENT_ID_SEP)
            read_from_disk(path, deserializers_patterns, {})
        return self

    def to_disk(
        self: SpaczzRuler,
        path: Union[str, Path],
        *,
        exclude: Iterable[str] = simple_frozen_list,
    ) -> None:
        """Save the spaczz ruler patterns to a directory.

        The patterns will be saved as newline-delimited JSON (JSONL).

        Args:
            path: The JSONL file to save.
            exclude: For spaCy consistency.

        Example:
            >>> import os
            >>> import tempfile
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            >>>     ruler.to_disk(f"{tmpdir}/ruler")
            >>>     isdir = os.path.isdir(f"{tmpdir}/ruler")
            >>> isdir
            True
        """
        path = ensure_path(path)
        cfg = {
            "overwrite": self.overwrite,
            "defaults": self.defaults,
            "ent_id_sep": self.ent_id_sep,
        }
        serializers = {
            "patterns": lambda p: srsly.write_jsonl(
                p.with_suffix(".jsonl"), self.patterns
            ),
            "cfg": lambda p: srsly.write_json(p, cfg),
        }
        if path.suffix == ".jsonl":  # user wants to save only JSONL
            srsly.write_jsonl(path, self.patterns)
        else:
            write_to_disk(path, serializers, {})

    def _add_patterns(
        self: SpaczzRuler,
        fuzzy_patterns: List[Dict[str, Any]],
        regex_patterns: List[Dict[str, Any]],
        token_patterns: List[Dict[str, Any]],
    ) -> None:
        """Helper function for add_patterns."""
        for entry in fuzzy_patterns + regex_patterns + token_patterns:
            label = entry["label"]
            if "id" in entry:
                ent_label = label
                label = self._create_label(label, entry["id"])
                self._ent_ids[label] = (ent_label, entry["id"])
            pattern = entry["pattern"]
            if isinstance(pattern, Doc):
                self.fuzzy_patterns[label]["patterns"].append(pattern)
                self.fuzzy_patterns[label]["kwargs"].append(entry["kwargs"])
            elif isinstance(pattern, str):
                self.regex_patterns[label]["patterns"].append(pattern)
                self.regex_patterns[label]["kwargs"].append(entry["kwargs"])
            elif isinstance(pattern, list):
                self.token_patterns[label].append(pattern)
            else:
                raise ValueError(
                    (
                        "One or more patterns do not conform",
                        "to spaczz pattern structure:",
                        "{label (str), pattern (str or list), type (str),",
                        "optional kwargs (dict[str, Any]),",
                        "and optional id (str)}.",
                    )
                )
        for label, patterns in self.fuzzy_patterns.items():
            self.fuzzy_matcher.add(label, patterns["patterns"], patterns["kwargs"])
        for label, patterns in self.regex_patterns.items():
            self.regex_matcher.add(label, patterns["patterns"], patterns["kwargs"])
        for label, _token_patterns in self.token_patterns.items():
            self.token_matcher.add(label, _token_patterns)

    def _create_label(self: SpaczzRuler, label: str, ent_id: Union[str, None]) -> str:
        """Join Entity label with ent_id if the pattern has an id attribute.

        Args:
            label: The entity label.
            ent_id: The optional entity id.

        Returns:
            The label and ent_id joined with configured ent_id_sep.
        """
        if isinstance(ent_id, str):
            label = "{}{}{}".format(label, self.ent_id_sep, ent_id)
        return label

    def _split_label(self: SpaczzRuler, label: str) -> Tuple[str, Union[str, None]]:
        """Split Entity label into ent_label and ent_id if it contains self.ent_id_sep.

        Args:
            label: The value of label in a pattern entry

        Returns:
            The separated ent_label and optional ent_id.
        """
        if self.ent_id_sep in label:
            ent_label, ent_id = label.rsplit(self.ent_id_sep, 1)
            return ent_label, ent_id
        else:
            ent_label = label
            return ent_label, None

    @staticmethod
    def _filter_overlapping_matches(
        matches: List[Tuple[str, int, int]],
        lookup: DefaultDict[str, Dict[Tuple[str, int, int], Any]],
    ) -> Tuple[
        List[Tuple[str, int, int]], DefaultDict[str, Dict[Tuple[str, int, int], Any]]
    ]:
        """Prevents multiple match spans from overlapping.

        Expects matches to be pre-sorted by matcher priority,
        with each matcher's matches being pre-sorted by descending length,
        then ascending start index, then descending match score
        If more than one match span includes the same tokens
        the first of these match spans in matches is kept.

        It also removes non-kept matches from the lookup dict as well.

        Args:
            matches: List of match span tuples
                (match_id, start_index, end_index).
            lookup: Match ratio, count and detail values in
                a `defaultdict(dict)`.

        Returns:
            The filtered list of match span tuples.
        """
        filtered_matches: List[Tuple[str, int, int]] = []
        for match in matches:
            if not set(range(match[1], match[2])).intersection(
                chain(*[set(range(n[1], n[2])) for n in filtered_matches])
            ):
                filtered_matches.append(match)
                if match in lookup["ratios"]:
                    _ = lookup["counts"].pop(match, None)
                    _ = lookup["details"].pop(match, None)
                elif match in lookup["counts"]:
                    _ = lookup["details"].pop(match, None)
        return filtered_matches, lookup

    @staticmethod
    def _update_custom_attrs(
        span: Span,
        match_id: str,
        lookup: DefaultDict[str, Dict[Tuple[str, int, int], Any]],
    ) -> Span:
        """Update custom attributes for matches."""
        ratio = lookup["ratios"].get((match_id, span.start, span.end))
        counts = lookup["counts"].get((match_id, span.start, span.end))
        details = lookup["details"].get((match_id, span.start, span.end))
        for token in span:
            token._.spaczz_token = True
            if ratio:
                token._.spaczz_ratio = ratio
                token._.spaczz_type = "fuzzy"
            elif counts:
                token._.spaczz_counts = counts
                token._.spaczz_type = "regex"
            elif details:
                token._.spaczz_details = details
                token._.spaczz_type = "token"
        return span



"""Module for FuzzySearcher: fuzzy matching in spaCy `Doc` objects."""
from __future__ import annotations

from typing import Any, Union

from spacy.tokens import Doc, Span, Token
from spacy.vocab import Vocab

from . import _PhraseSearcher
from ..process import FuzzyFuncs


class FuzzySearcher(_PhraseSearcher):
    """Class for fuzzy searching/matching in spacy `Doc` objects.

    Fuzzy searching is done on the token level.
    The class provides methods for finding the best fuzzy match
    span in a `Doc`, the n best fuzzy matched spans in a `Doc`,
    and fuzzy matching between any two given spaCy containers
    (`Doc`, `Span`, `Token`).

    Fuzzy matching is currently provided by rapidfuzz and the searcher
    provides access to all rapidfuzz matchers with default settings.

    Attributes:
        vocab (Vocab): The shared vocabulary.
            Included for consistency and potential future-state.
        _fuzzy_funcs (FuzzyFuncs):
            Container class housing fuzzy matching functions.
            Functions are accessible via the classes `get()` method
            by their given key name. All rapidfuzz matching functions
            with default settings are available:
            "simple" = `ratio`
            "partial" = `partial_ratio`
            "token_set" = `token_set_ratio`
            "token_sort" = `token_sort_ratio`
            "partial_token_set" = `partial_token_set_ratio`
            "partial_token_sort" = `partial_token_sort_ratio`
            "quick" = `QRatio`
            "weighted" = `WRatio`
            "quick_lev" = `quick_lev_ratio`
    """

    def __init__(self: FuzzySearcher, vocab: Vocab) -> None:
        """Initializes a fuzzy searcher.

        Args:
            vocab: A spaCy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are mostly pure-Python
                currently and do not share vocabulary
                with spaCy pipelines.
        """
        super().__init__(vocab=vocab)
        self._fuzzy_funcs: FuzzyFuncs = FuzzyFuncs(match_type="phrase")

    def compare(
        self: FuzzySearcher,
        query: Union[Doc, Span, Token],
        reference: Union[Doc, Span, Token],
        ignore_case: bool = True,
        fuzzy_func: str = "simple",
        *args: Any,
        **kwargs: Any,
    ) -> int:
        """Peforms fuzzy matching between two spaCy container objects.

        Applies the given fuzzy matching algorithm (`fuzzy_func`)
        to two spacy containers (`Doc`, `Span`, `Token`)
        and returns the resulting fuzzy ratio.

        Args:
            query: First container for comparison.
            reference: Second container for comparison.
            ignore_case: Whether to lower-case `query` and `reference`
                before comparison or not. Default is `True`.
            fuzzy_func: Key name of fuzzy matching function to use.
                All rapidfuzz matching functions with default settings
                are available:
                "simple" = `ratio`
                "partial" = `partial_ratio`
                "token_set" = `token_set_ratio`
                "token_sort" = `token_sort_ratio`
                "partial_token_set" = `partial_token_set_ratio`
                "partial_token_sort" = `partial_token_sort_ratio`
                "quick" = `QRatio`
                "weighted" = `WRatio`
                "quick_lev" = `quick_lev_ratio`
                Default is `"simple"`.
            *args: Overflow for child positional arguments.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            The fuzzy ratio between `query` and `reference` as an `int`.

        Example:
            >>> import spacy
            >>> from spaczz.search import FuzzySearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = FuzzySearcher(nlp.vocab)
            >>> searcher.compare(nlp("spaczz"), nlp("spacy"))
            73
        """
        if ignore_case:
            query_text = query.text.lower()
            reference_text = reference.text.lower()
        else:
            query_text = query.text
            reference_text = reference.text
        return round(self._fuzzy_funcs.get(fuzzy_func)(query_text, reference_text))



"""Module for various doc/text processing classes/functions."""
from __future__ import annotations

from typing import Callable, Dict

from rapidfuzz import fuzz
from spacy.tokens import Doc


def map_chars_to_tokens(doc: Doc) -> Dict[int, int]:
    """Maps characters in a `Doc` object to tokens."""
    chars_to_tokens = {}
    for token in doc:
        for i in range(token.idx, token.idx + len(token.text)):
            chars_to_tokens[i] = token.i
    return chars_to_tokens


class FuzzyFuncs:
    """Container class housing fuzzy matching functions.

    Functions are accessible via the classes `get()` method
    by their given key name. All rapidfuzz matching functions
    with default settings are available.

    Attributes:
        match_type (str): Whether the fuzzy matching functions
            should support multi-token strings ("phrase") or
            only single-token strings ("token").
        _fuzzy_funcs (dict[str, Callable[[str, str], int]]):
            The available fuzzy matching functions:
            "simple" = `ratio`
            "partial" = `partial_ratio`
            "token_set" = `token_set_ratio`
            "token_sort" = `token_sort_ratio`
            "partial_token_set" = `partial_token_set_ratio`
            "partial_token_sort" = `partial_token_sort_ratio`
            "quick" = `QRatio`
            "weighted" = `WRatio`
            "quick_lev" = `quick_lev_ratio`
            This is limited to "simple", "quick", and "quick_lev"
            if match_type = "token".
    """

    def __init__(self: FuzzyFuncs, match_type: str = "phrase") -> None:
        """Initializes a `FuzzyFuncs` container.

        Args:
            match_type: Whether the fuzzy matching functions
                support multi-token strings ("phrase") or
                only single-token strings ("token").

        Raises:
            ValueError: If match_type is not "phrase" or "token".
        """
        self.match_type = match_type
        if match_type == "phrase":
            self._fuzzy_funcs: Dict[str, Callable[[str, str], int]] = {
                "simple": fuzz.ratio,
                "partial": fuzz.partial_ratio,
                "token_set": fuzz.token_set_ratio,
                "token_sort": fuzz.token_sort_ratio,
                "partial_token_set": fuzz.partial_token_set_ratio,
                "partial_token_sort": fuzz.partial_token_sort_ratio,
                "quick": fuzz.QRatio,
                "weighted": fuzz.WRatio,
                "token": fuzz.token_ratio,
                "partial_token": fuzz.partial_token_ratio,
            }
        elif match_type == "token":
            self._fuzzy_funcs = {
                "simple": fuzz.ratio,
                "quick": fuzz.QRatio,
            }
        else:
            raise ValueError("match_type must be either 'phrase' or 'token'.")

    def get(self: FuzzyFuncs, fuzzy_func: str) -> Callable[[str, str], float]:
        """Returns a fuzzy matching function based on it's key name.

        Args:
            fuzzy_func: Key name of the fuzzy matching function.

        Returns:
            A fuzzy matching function.

        Raises:
            ValueError: The fuzzy function was not a valid key name.

        Example:
            >>> import spacy
            >>> from spaczz.process import FuzzyFuncs
            >>> ff = FuzzyFuncs()
            >>> simple = ff.get("simple")
            >>> simple("hi", "hi")
            100.0
        """
        try:
            return self._fuzzy_funcs[fuzzy_func.lower()]
        except KeyError:
            raise ValueError(
                (
                    f"No fuzzy {self.match_type} matching function",
                    f"called: {fuzzy_func}.",
                    "Matching function must be in the following (case insensitive):",
                    f"{list(self._fuzzy_funcs.keys())}",
                )
            )



"""Custom spaCy attributes for spaczz."""
from __future__ import annotations

from typing import Iterable, Optional, Set, Tuple, Type
import warnings

from spacy.tokens import Doc, Span, Token

from .exceptions import AttrOverwriteWarning, SpaczzSpanDeprecation


class SpaczzAttrs:
    """Adds spaczz custom attributes to spacy."""

    _initialized = False

    @classmethod
    def initialize(cls: Type[SpaczzAttrs]) -> None:
        """Initializes and registers custom attributes."""
        if not cls._initialized:
            try:
                Token.set_extension("spaczz_token", default=False)
                Token.set_extension("spaczz_type", default=None)
                Token.set_extension("spaczz_ratio", default=None)
                Token.set_extension("spaczz_counts", default=None)
                Token.set_extension("spaczz_details", default=None)

                Span.set_extension("spaczz_span", getter=cls.get_spaczz_span)
                Span.set_extension("spaczz_ent", getter=cls.get_spaczz_ent)
                Span.set_extension("spaczz_type", getter=cls.get_span_type)
                Span.set_extension("spaczz_types", getter=cls.get_span_types)
                Span.set_extension("spaczz_ratio", getter=cls.get_ratio)
                Span.set_extension("spaczz_counts", getter=cls.get_counts)
                Span.set_extension("spaczz_details", getter=cls.get_details)

                Doc.set_extension("spaczz_doc", getter=cls.get_spaczz_doc)
                Doc.set_extension("spaczz_types", getter=cls.get_doc_types)
                cls._initialized = True
            except ValueError:
                warnings.warn(
                    """One or more spaczz custom extensions has already been registered.
                    These are being force overwritten. Please avoid defining personal,
                    custom extensions prepended with "spaczz_".
                """,
                    AttrOverwriteWarning,
                )
                Token.set_extension("spaczz_token", default=False, force=True)
                Token.set_extension("spaczz_type", default=None, force=True)
                Token.set_extension("spaczz_ratio", default=None, force=True)
                Token.set_extension("spaczz_counts", default=None, force=True)

                Span.set_extension(
                    "spaczz_span", getter=cls.get_spaczz_span, force=True
                )
                Span.set_extension("spaczz_type", getter=cls.get_span_type, force=True)
                Span.set_extension(
                    "spaczz_types", getter=cls.get_span_types, force=True
                )
                Span.set_extension("spaczz_ratio", getter=cls.get_ratio, force=True)
                Span.set_extension("spaczz_counts", getter=cls.get_counts, force=True)

                Doc.set_extension("spaczz_doc", getter=cls.get_spaczz_doc, force=True)
                Doc.set_extension("spaczz_types", getter=cls.get_doc_types, force=True)

    @staticmethod
    def get_spaczz_span(span: Span) -> bool:
        """Getter for spaczz_span `Span` attribute."""
        warnings.warn(
            """spaczz_span is deprecated.
        Use spaczz_ent instead.""",
            SpaczzSpanDeprecation,
        )
        return all([token._.spaczz_token for token in span])

    @staticmethod
    def get_spaczz_ent(span: Span) -> bool:
        """Getter for spaczz_ent `Span` attribute."""
        return all([token._.spaczz_token for token in span])

    @classmethod
    def get_span_type(cls: Type[SpaczzAttrs], span: Span) -> Optional[str]:
        """Getter for spaczz_type `Span` attribute."""
        if cls._all_equal([token._.spaczz_type for token in span]):
            return span[0]._.spaczz_type
        else:
            return None

    @staticmethod
    def get_span_types(span: Span) -> Set[str]:
        """Getter for spaczz_types `Span` attribute."""
        types = [token._.spaczz_type for token in span if token._.spaczz_type]
        return set(types)

    @classmethod
    def get_ratio(cls: Type[SpaczzAttrs], span: Span) -> Optional[int]:
        """Getter for spaczz_ratio `Span` attribute."""
        if cls._all_equal([token._.spaczz_ratio for token in span]):
            return span[0]._.spaczz_ratio
        else:
            return None

    @classmethod
    def get_counts(
        cls: Type[SpaczzAttrs], span: Span
    ) -> Optional[Tuple[int, int, int]]:
        """Getter for spaczz_counts `Span` attribute."""
        if cls._all_equal([token._.spaczz_counts for token in span]):
            return span[0]._.spaczz_counts
        else:
            return None

    @classmethod
    def get_details(cls: Type[SpaczzAttrs], span: Span) -> Optional[int]:
        """Getter for current placeholder spaczz_details `Span` attribute."""
        if cls._all_equal([token._.spaczz_details for token in span]):
            return span[0]._.spaczz_details
        else:
            return None

    @staticmethod
    def get_spaczz_doc(doc: Doc) -> bool:
        """Getter for spaczz_doc `Doc` attribute."""
        return any([token._.spaczz_token for token in doc])

    @staticmethod
    def get_doc_types(doc: Doc) -> Set[str]:
        """Getter for spaczz_types `Doc` attribute."""
        types = [token._.spaczz_type for token in doc if token._.spaczz_type]
        return set(types)

    @staticmethod
    def _all_equal(iterable: Iterable) -> bool:
        """Tests if all elements of iterable are equal."""
        iterator = iter(iterable)
        try:
            first = next(iterator)
        except StopIteration:
            return True
        return all(first == rest for rest in iterator)



"""Module for TokenMatcher with an API semi-analogous to spaCy's Matcher."""
from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
)
import warnings

from spacy.matcher import Matcher
from spacy.tokens import Doc
from spacy.vocab import Vocab

from ..exceptions import PipeDeprecation
from ..search import TokenSearcher


class TokenMatcher:
    """spaCy-like token matcher for finding flexible matches in `Doc` objects.

    Matches added patterns against the `Doc` object it is called on.
    Accepts labeled patterns in the form of lists of dictionaries
    where each list describes an individual pattern and each
    dictionary describes an individual token.

    Uses extended spaCy token matching patterns.
    "FUZZY" and "FREGEX" are the two additional spaCy token pattern options.

    For example:
        {"TEXT": {"FREGEX": "(database){e<=1}"}},
        {"LOWER": {"FUZZY": "access", "MIN_R": 85, "FUZZY_FUNC": "quick_lev"}}

    Make sure to use uppercase dictionary keys in patterns.

    Attributes:
        defaults: Keyword arguments to be used as default matching settings.
            See `TokenSearcher.match()` documentation for details.
        name: Class attribute - the name of the matcher.
        type: The kind of matcher object.
        _callbacks:
            On match functions to modify `Doc` objects passed to the matcher.
            Can make use of the matches identified.
        _patterns:
            Patterns added to the matcher.
    """

    name = "token_matcher"

    def __init__(self: TokenMatcher, vocab: Vocab, **defaults: Any) -> None:
        """Initializes the base phrase matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are currently pure
                Python and do not share vocabulary
                with spaCy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings.
                These arguments will become the new defaults for matching.
                See `TokenSearcher.match()` documentation for details.
        """
        self.defaults = defaults
        self.type = "token"
        self._callbacks: Dict[str, TokenCallback] = {}
        self._patterns: DefaultDict[str, List[List[Dict[str, Any]]]] = DefaultDict(list)
        self._searcher = TokenSearcher(vocab=vocab)

    def __call__(self: TokenMatcher, doc: Doc) -> List[Tuple[str, int, int, None]]:
        """Find all sequences matching the supplied patterns in the doc.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of (key, start, end, None) tuples, describing the matches.
            The final None is a placeholder for future match details.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> doc = nlp("Rdley Scot was the director of Alien.")
            >>> matcher.add("NAME", [
                [{"TEXT": {"FUZZY": "Ridley"}},
                {"TEXT": {"FUZZY": "Scott"}}]
                ])
            >>> matcher(doc)
            [('NAME', 0, 2, None)]
        """
        mapped_patterns = defaultdict(list)
        matcher = Matcher(self.vocab)
        for label, patterns in self._patterns.items():
            for pattern in patterns:
                mapped_patterns[label].extend(
                    _spacyfy(
                        self._searcher.match(doc, pattern, **self.defaults),
                        pattern,
                    )
                )
        for label in mapped_patterns.keys():
            matcher.add(label, mapped_patterns[label])
        matches = matcher(doc)
        if matches:
            extended_matches = [
                (self.vocab.strings[match_id], start, end, None)
                for match_id, start, end in matches
            ]
            for i, (label, _start, _end, _details) in enumerate(extended_matches):
                on_match = self._callbacks.get(label)
                if on_match:
                    on_match(self, doc, i, extended_matches)
            return extended_matches
        else:
            return []

    def __contains__(self: TokenMatcher, label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return label in self._patterns

    def __len__(self: TokenMatcher) -> int:
        """The number of labels added to the matcher."""
        return len(self._patterns)

    def __reduce__(
        self: TokenMatcher,
    ) -> Tuple[Any, Any]:  # Precisely typing this would be really long.
        """Interface for pickling the matcher."""
        data = (
            self.__class__,
            self.vocab,
            self._patterns,
            self._callbacks,
            self.defaults,
        )
        return (unpickle_matcher, data)

    @property
    def labels(self: TokenMatcher) -> Tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique string labels as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> matcher.labels
            ('AUTHOR',)
        """
        return tuple(self._patterns.keys())

    @property
    def patterns(self: TokenMatcher) -> List[Dict[str, Any]]:
        """Get all patterns that were added to the matcher.

        Returns:
            The original patterns, one dictionary for each combination.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> matcher.patterns == [
                {
                    "label": "AUTHOR",
                    "pattern": [{"TEXT": {"FUZZY": "Kerouac"}}],
                    "type": "token",
                    },
                    ]
            True
        """
        all_patterns = []
        for label, patterns in self._patterns.items():
            for pattern in patterns:
                p = {"label": label, "pattern": pattern, "type": self.type}
                all_patterns.append(p)
        return all_patterns

    @property
    def vocab(self: TokenMatcher) -> Vocab:
        """Returns the spaCy `Vocab` object utilized."""
        return self._searcher.vocab

    def add(
        self: TokenMatcher,
        label: str,
        patterns: List[List[Dict[str, Any]]],
        on_match: TokenCallback = None,
    ) -> None:
        """Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of dictionary lists where each dictionary
        list represent an individual pattern and each dictionary represents
        an individual token.

        Uses extended spaCy token matching patterns.
        "FUZZY" and "FREGEX" are the two additional spaCy token pattern options.

        For example:
            {"TEXT": {"FREGEX": "(database){e<=1}"}},
            {"LOWER": {"FUZZY": "access", "MIN_R": 85, "FUZZY_FUNC": "quick_lev"}}

        Args:
            label: Name of the rule added to the matcher.
            patterns: List of dictionary lists that will be matched
                against the `Doc` object the matcher is called on.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Raises:
            TypeError: If patterns is not an iterable of `Doc` objects.
            ValueError: pattern cannot have zero tokens.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> "AUTHOR" in matcher
            True
        """
        for pattern in patterns:
            if len(pattern) == 0:
                raise ValueError("pattern cannot have zero tokens.")
            if isinstance(pattern, list):
                self._patterns[label].append(list(pattern))
            else:
                raise TypeError("Patterns must be lists of dictionaries.")
        self._callbacks[label] = on_match

    def remove(self: TokenMatcher, label: str) -> None:
        """Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Raises:
            ValueError: If label does not exist in the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> matcher.remove("AUTHOR")
            >>> "AUTHOR" in matcher
            False
        """
        try:
            del self._patterns[label]
            del self._callbacks[label]
        except KeyError:
            raise ValueError(
                f"The label: {label} does not exist within the matcher rules."
            )

    def pipe(
        self: TokenMatcher,
        stream: Iterable[Doc],
        batch_size: int = 1000,
        return_matches: bool = False,
        as_tuples: bool = False,
    ) -> Generator[Any, None, None]:
        """Match a stream of `Doc` objects, yielding them in turn.

        Deprecated as of spaCy v3.0 and spaczz v0.5.

        Args:
            stream: A stream of `Doc` objects.
            batch_size: Number of documents to accumulate into a working set.
                Default is `1000`.
            return_matches: Yield the match lists along with the docs,
                making results (doc, matches) tuples. Default is `False`.
            as_tuples: Interpret the input stream as (doc, context) tuples,
                and yield (result, context) tuples out.
                If both return_matches and as_tuples are `True`,
                the output will be a sequence of ((doc, matches), context) tuples.
                Default is `False`.

        Yields:
            `Doc` objects, in order.
        """
        warnings.warn(
            """As of spaczz v0.5 and spaCy v3.0, the matcher.pipe method
        is deprecated. If you need to match on a stream of documents,
        you can use nlp.pipe and call the matcher on each Doc object.""",
            PipeDeprecation,
        )
        if as_tuples:
            for doc, context in stream:
                matches = self(doc)
                if return_matches:
                    yield ((doc, matches), context)
                else:
                    yield (doc, context)
        else:
            for doc in stream:
                matches = self(doc)
                if return_matches:
                    yield (doc, matches)
                else:
                    yield doc


def _spacyfy(
    matches: List[List[Optional[Tuple[str, str]]]], pattern: List[Dict[str, Any]]
) -> List[List[Dict[str, Any]]]:
    """Turns token searcher matches into spaCy `Matcher` compatible patterns."""
    new_patterns = []
    if matches:
        for match in matches:
            new_pattern = deepcopy(pattern)
            for i, token in enumerate(match):
                if token:
                    del new_pattern[i][token[0]]
                    new_pattern[i]["TEXT"] = token[1]
            new_patterns.append(new_pattern)
    return new_patterns


TokenCallback = Optional[
    Callable[[TokenMatcher, Doc, int, List[Tuple[str, int, int, None]]], None]
]


def unpickle_matcher(
    matcher: Type[TokenMatcher],
    vocab: Vocab,
    patterns: DefaultDict[str, List[List[Dict[str, Any]]]],
    callbacks: Dict[str, TokenCallback],
    defaults: Any,
) -> Any:
    """Will return a matcher from pickle protocol."""
    matcher_instance = matcher(vocab, **defaults)
    for key, specs in patterns.items():
        callback = callbacks.get(key)
        matcher_instance.add(key, specs, on_match=callback)
    return matcher_instance



"""Module for spaCy v2 compatible SpaczzRuler."""
from __future__ import annotations

from collections import defaultdict, OrderedDict
from itertools import chain
from pathlib import Path
from typing import Any, DefaultDict, Dict, Iterable, List, Optional, Set, Tuple, Union
import warnings

try:
    from spacy.language import component, Language
except ImportError:  # pragma: no cover
    raise ImportError(
        (
            "Trying to import spaCy v2 compatible SpaczzRuler from spaCy v3.",
            "Please downgrade or use the SpaczzRuler in _spaczzruler",
        )
    )
from spacy.tokens import Doc, Span
import srsly

from ..exceptions import PatternTypeWarning
from ..matcher import FuzzyMatcher, RegexMatcher, TokenMatcher
from ..util import ensure_path, nest_defaultdict, read_from_disk, write_to_disk


DEFAULT_ENT_ID_SEP = "||"


def make_spaczz_ruler(nlp: Any, **cfg: Any) -> SpaczzRuler:
    """Factory method for creating a `SpaczzRuler`."""
    return SpaczzRuler(nlp, **cfg)


@component("spaczz_ruler", assigns=["doc.ents", "token.ent_type", "token.ent_iob"])
class SpaczzRuler:
    """The `SpaczzRuler` adds fuzzy and multi-token regex matches to spaCy `Doc.ents`.

    It can be combined with other spaCy NER components like the statistical
    EntityRecognizer and/or the EntityRuler to boost accuracy.
    After initialization, the component is typically added to the pipeline
    using `nlp.add_pipe`.

    Attributes:
        nlp: The shared nlp object to pass the vocab to the matchers
            (not currently used by spaczz matchers) and process fuzzy patterns.
        fuzzy_patterns:
            Patterns added to the fuzzy matcher.
        regex_patterns:
            Patterns added to the regex matcher.
        token_patterns:
            Patterns added to the token matcher
        fuzzy_matcher: The `FuzzyMatcher` instance
            the spaczz ruler will use for fuzzy phrase matching.
        regex_matcher: The `RegexMatcher` instance
            the spaczz ruler will use for regex phrase matching.
        token_matcher: The `TokenMatcher` instance
            the spaczz ruler will use for token matching.
        defaults: Default matching settings for their respective matchers.
    """

    name = "spaczz_ruler"

    def __init__(self: SpaczzRuler, nlp: Language, **cfg: Any) -> None:
        """Initialize the spaczz ruler with a Language object and cfg parameters.

        All spaczz ruler cfg parameters are prepended with `"spaczz_"`.
        If `spaczz_patterns` is supplied here, it needs to be a list of spaczz patterns:
        dictionaries with `"label"`, `"pattern"`, and "type" keys.
        If the patterns are fuzzy or regex phrase patterns they can
        include the optional `"kwargs"` keys.

        For example, a fuzzy phrase pattern:
        `{'label': 'ORG', 'pattern': 'Apple',
        'type': 'fuzzy', 'kwargs': {'min_r2': 90}}`

        Or, a token pattern:
        `{'label': 'ORG', 'pattern': [{'TEXT': {'FUZZY': 'Apple'}}], 'type': 'token'}`

        Args:
            nlp: The shared `Language` object to pass the vocab to the matchers
                and process fuzzy patterns.
            **cfg: Other config parameters. The SpaczzRuler makes heavy use
                of cfg to pass additional parameters down to the matchers.
                spaczz config parameters start with "spaczz_" to keep them
                from colliding with other cfg components.
                SpaczzRuler cfg components include (with "spaczz_" prepended to them):
                    `overwrite_ents` (bool): Whether to overwrite exisiting Doc.ents
                        with new matches. Default is False.
                    `ent_id_sep` (str): String to separate entity labels and ids on.
                    `fuzzy_defaults` (dict[str, Any]): Modified default parameters to
                        use with the fuzzy matcher. Default is an empty dictionary -
                        utilizing defaults.
                    `regex_defaults` (dict[str, Any]): Modified default parameters to
                        use with the regex matcher. Default is an empty dictionary -
                        utilizing defaults.
                    `token_defaults` (dict[str, Any]): Modified default parameters to
                        use with the spaczz matcher. Default is an empty dictionary -
                        utilizing defaults.
                    `regex_config` (Union[str, RegexConfig]): Should largely be ignored
                        as an artifact of an old spaczz design pattern.
                        Will likely be updated in the future.
                        Default is `"default"`.
                    `patterns` (Iterable[dict[str, Any]]): Patterns to initialize
                        the ruler with. Default is None.
                If the spaczz ruler is loaded as part of a model pipeline,
                cfg will include all keyword arguments passed to `spacy.load`.

        Raises:
            TypeError: If spaczz_{name}_defaults passed are not dictionaries.
        """
        self.nlp = nlp
        self.fuzzy_patterns: DefaultDict[str, DefaultDict[str, Any]] = nest_defaultdict(
            list, 2
        )
        self.regex_patterns: DefaultDict[str, DefaultDict[str, Any]] = nest_defaultdict(
            list, 2
        )
        self.token_patterns: DefaultDict[str, List[List[Dict[str, Any]]]] = defaultdict(
            list
        )
        self.ent_id_sep = cfg.get("spaczz_ent_id_sep", DEFAULT_ENT_ID_SEP)
        self._ent_ids: DefaultDict[Any, Any] = defaultdict(dict)
        self.overwrite = cfg.get("spaczz_overwrite_ents", False)
        default_names = (
            ("spaczz_fuzzy_defaults", "fuzzy_defaults"),
            ("spaczz_regex_defaults", "regex_defaults"),
            ("spaczz_token_defaults", "token_defaults"),
        )
        self.defaults = {}
        for cfg_name, real_name in default_names:
            if cfg_name in cfg:
                if isinstance(cfg[cfg_name], dict):
                    self.defaults[real_name] = cfg[cfg_name]
                else:
                    raise TypeError(
                        (
                            "Defaults must be a dictionary of keyword arguments,",
                            f"not {type(cfg[cfg_name])}.",
                        )
                    )
        self.fuzzy_matcher = FuzzyMatcher(
            nlp.vocab,
            **self.defaults.get("fuzzy_defaults", {}),
        )
        self.regex_matcher = RegexMatcher(
            nlp.vocab,
            cfg.get("spaczz_regex_config", "default"),
            **self.defaults.get("regex_defaults", {}),
        )
        self.token_matcher = TokenMatcher(
            nlp.vocab, **self.defaults.get("token_defaults", {})
        )
        patterns = cfg.get("spaczz_patterns")
        if patterns is not None:
            self.add_patterns(patterns)

    def __call__(self: SpaczzRuler, doc: Doc) -> Doc:
        """Find matches in document and add them as entities.

        Args:
            doc: The Doc object in the pipeline.

        Returns:
            The Doc with added entities, if available.
        """
        matches, lookup = self.match(doc)
        self.set_annotations(doc, matches, lookup)
        return doc

    def __contains__(self: SpaczzRuler, label: str) -> bool:
        """Whether a label is present in the patterns."""
        return (
            label in self.fuzzy_patterns
            or label in self.regex_patterns
            or label in self.token_patterns
        )

    def __len__(self: SpaczzRuler) -> int:
        """The number of all patterns added to the ruler."""
        n_fuzzy_patterns = sum(len(p["patterns"]) for p in self.fuzzy_patterns.values())
        n_regex_patterns = sum(len(p["patterns"]) for p in self.regex_patterns.values())
        n_token_patterns = sum(len(p) for p in self.token_patterns.values())
        return n_fuzzy_patterns + n_regex_patterns + n_token_patterns

    @property
    def ent_ids(self: SpaczzRuler) -> Tuple[Optional[str], ...]:
        """All entity ids present in the match patterns id properties.

        Returns:
            The unique string entity ids as a tuple.
        """
        keys = set(self.fuzzy_patterns.keys())
        keys.update(self.regex_patterns.keys())
        keys.update(self.token_patterns.keys())
        all_ent_ids = set()

        for k in keys:
            if self.ent_id_sep in k:
                _, ent_id = self._split_label(k)
                all_ent_ids.add(ent_id)
        all_ent_ids_tuple = tuple(all_ent_ids)
        return all_ent_ids_tuple

    @property
    def labels(self: SpaczzRuler) -> Tuple[str, ...]:
        """All labels present in the ruler.

        Returns:
            The unique string labels as a tuple.
        """
        keys = set(self.fuzzy_patterns.keys())
        keys.update(self.regex_patterns.keys())
        keys.update(self.token_patterns.keys())
        all_labels = set()
        for k in keys:
            if self.ent_id_sep in k:
                label, _ = self._split_label(k)
                all_labels.add(label)
            else:
                all_labels.add(k)
        return tuple(all_labels)

    @property
    def patterns(self: SpaczzRuler) -> List[Dict[str, Any]]:
        """Get all patterns and kwargs that were added to the ruler.

        Returns:
            The original patterns and kwargs, one dictionary for each combination.
        """
        all_patterns = []
        for label, fuzzy_patterns in self.fuzzy_patterns.items():
            for fuzzy_pattern, fuzzy_kwargs in zip(
                fuzzy_patterns["patterns"], fuzzy_patterns["kwargs"]
            ):
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": fuzzy_pattern.text, "type": "fuzzy"}
                if fuzzy_kwargs:
                    p["kwargs"] = fuzzy_kwargs
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        for label, regex_patterns in self.regex_patterns.items():
            for regex_pattern, regex_kwargs in zip(
                regex_patterns["patterns"], regex_patterns["kwargs"]
            ):
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": regex_pattern, "type": "regex"}
                if regex_kwargs:
                    p["kwargs"] = regex_kwargs
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        for label, token_patterns in self.token_patterns.items():
            for token_pattern in token_patterns:
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": token_pattern, "type": "token"}
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        return all_patterns

    def add_patterns(
        self: SpaczzRuler,
        patterns: Iterable[Dict[str, Any]],
    ) -> None:
        """Add patterns to the ruler.

        A pattern must be a spaczz pattern:
        `{label (str), pattern (str or list), type (str),
        optional kwargs (dict[str, Any]), and optional id (str)}`.

        For example, a fuzzy phrase pattern:
        `{'label': 'ORG', 'pattern': 'Apple',
        'type': 'fuzzy', 'kwargs': {'min_r2': 90}}`

        Or, a token pattern:
        `{'label': 'ORG', 'pattern': [{'TEXT': {'FUZZY': 'Apple'}}], 'type': 'token'}`

        To utilize regex flags, use inline flags.
        Kwarg details to be updated.

        Args:
            patterns: The spaczz patterns to add.

        Raises:
            TypeError: If patterns is not an iterable of dictionaries.
            ValueError: If one or more patterns do not conform
                the spaczz pattern structure.
        """
        # disable the nlp components after this one
        # in case they haven't been initialized / deserialised yet
        try:
            current_index = self.nlp.pipe_names.index(self.name)
            subsequent_pipes = [
                pipe for pipe in self.nlp.pipe_names[current_index + 1 :]
            ]
        except ValueError:
            subsequent_pipes = []

        with self.nlp.disable_pipes(subsequent_pipes):
            token_patterns = []
            fuzzy_pattern_labels = []
            fuzzy_pattern_texts = []
            fuzzy_pattern_kwargs = []
            fuzzy_pattern_ids = []
            regex_pattern_labels = []
            regex_pattern_texts = []
            regex_pattern_kwargs = []
            regex_pattern_ids = []

            for entry in patterns:
                try:
                    if isinstance(entry, dict):
                        if entry["type"] == "fuzzy":
                            fuzzy_pattern_labels.append(entry["label"])
                            fuzzy_pattern_texts.append(entry["pattern"])
                            fuzzy_pattern_kwargs.append(entry.get("kwargs", {}))
                            fuzzy_pattern_ids.append(entry.get("id"))
                        elif entry["type"] == "regex":
                            regex_pattern_labels.append(entry["label"])
                            regex_pattern_texts.append(entry["pattern"])
                            regex_pattern_kwargs.append(entry.get("kwargs", {}))
                            regex_pattern_ids.append(entry.get("id"))
                        elif entry["type"] == "token":
                            token_patterns.append(entry)
                        else:
                            warnings.warn(
                                f"""Spaczz pattern "type" must be "fuzzy", "regex",
                                or "token", not {entry["type"]}. Skipping this pattern.
                                """,
                                PatternTypeWarning,
                            )
                    else:
                        raise TypeError(
                            ("Patterns must either be an iterable of dicts.")
                        )
                except KeyError:
                    raise ValueError(
                        (
                            "One or more patterns do not conform",
                            "to spaczz pattern structure: ",
                            "{label (str), pattern (str or list), type (str),",
                            "optional kwargs (dict[str, Any]),",
                            "and optional id (str)}.",
                        )
                    )

            fuzzy_patterns = []
            for label, pattern, kwargs, ent_id in zip(
                fuzzy_pattern_labels,
                self.nlp.pipe(fuzzy_pattern_texts),
                fuzzy_pattern_kwargs,
                fuzzy_pattern_ids,
            ):
                fuzzy_pattern = {
                    "label": label,
                    "pattern": pattern,
                    "kwargs": kwargs,
                    "type": "fuzzy",
                }
                if ent_id:
                    fuzzy_pattern["id"] = ent_id
                fuzzy_patterns.append(fuzzy_pattern)

            regex_patterns = []
            for label, pattern, kwargs, ent_id in zip(
                regex_pattern_labels,
                regex_pattern_texts,
                regex_pattern_kwargs,
                regex_pattern_ids,
            ):
                regex_pattern = {
                    "label": label,
                    "pattern": pattern,
                    "kwargs": kwargs,
                    "type": "regex",
                }
                if ent_id:
                    regex_pattern["id"] = ent_id
                regex_patterns.append(regex_pattern)

            self._add_patterns(fuzzy_patterns, regex_patterns, token_patterns)

    def match(
        self: SpaczzRuler, doc: Doc
    ) -> Tuple[
        List[Tuple[str, int, int]],
        DefaultDict[str, Dict[Tuple[str, int, int], Any]],
    ]:
        """Used in call to find matches in a doc."""
        fuzzy_matches = []
        lookup: DefaultDict[str, Dict[Tuple[str, int, int], Any]] = defaultdict(dict)
        for fuzzy_match in self.fuzzy_matcher(doc):
            current_ratio = fuzzy_match[3]
            best_ratio = lookup["ratios"].get(fuzzy_match[:3], 0)
            if current_ratio > best_ratio:
                fuzzy_matches.append(fuzzy_match[:3])
                lookup["ratios"][fuzzy_match[:3]] = current_ratio
        regex_matches = []
        for regex_match in self.regex_matcher(doc):
            current_counts = regex_match[3]
            best_counts = lookup["counts"].get(regex_match[:3])
            if not best_counts or sum(current_counts) < sum(best_counts):
                regex_matches.append(regex_match[:3])
                lookup["counts"][regex_match[:3]] = current_counts
        token_matches = []
        for token_match in self.token_matcher(doc):
            token_matches.append(token_match[:3])
            lookup["details"][token_match[:3]] = 1
        matches = fuzzy_matches + regex_matches + token_matches
        unique_matches, lookup = self._filter_overlapping_matches(matches, lookup)
        return unique_matches, lookup

    def set_annotations(
        self: SpaczzRuler,
        doc: Doc,
        matches: List[Tuple[str, int, int]],
        lookup: DefaultDict[
            str, Dict[Tuple[str, int, int], Union[int, Tuple[int, int, int]]]
        ],
    ) -> None:
        """Modify the document in place."""
        entities = list(doc.ents)
        new_entities = []
        seen_tokens: Set[int] = set()
        for match_id, start, end in matches:
            if any(t.ent_type for t in doc[start:end]) and not self.overwrite:
                continue
            # check for end - 1 here because boundaries are inclusive
            if start not in seen_tokens and end - 1 not in seen_tokens:
                if match_id in self._ent_ids:
                    label, ent_id = self._ent_ids[match_id]
                    span = Span(doc, start, end, label=label)
                    if ent_id:
                        for token in span:
                            token.ent_id_ = ent_id
                else:
                    span = Span(doc, start, end, label=match_id)
                span = self._update_custom_attrs(span, match_id, lookup)
                new_entities.append(span)
                entities = [
                    e for e in entities if not (e.start < end and e.end > start)
                ]
                seen_tokens.update(range(start, end))
        doc.ents = entities + new_entities

    def from_bytes(
        self: SpaczzRuler, patterns_bytes: bytes, **kwargs: Any
    ) -> SpaczzRuler:
        """Load the spaczz ruler from a bytestring.

        Args:
            patterns_bytes : The bytestring to load.
            **kwargs: Other config parameters, mostly for consistency.

        Returns:
            The loaded spaczz ruler.
        """
        cfg = srsly.msgpack_loads(patterns_bytes)
        if isinstance(cfg, dict):
            self.add_patterns(cfg.get("patterns", cfg))
            self.defaults = cfg.get("defaults", {})
            if self.defaults.get("fuzzy_defaults"):
                self.fuzzy_matcher = FuzzyMatcher(
                    self.nlp.vocab, **self.defaults["fuzzy_defaults"]
                )
            if self.defaults.get("regex_defaults"):
                self.regex_matcher = RegexMatcher(
                    self.nlp.vocab, **self.defaults["regex_defaults"]
                )
            if self.defaults.get("token_defaults"):
                self.token_matcher = TokenMatcher(
                    self.nlp.vocab, **self.defaults["token_defaults"]
                )
            self.overwrite = cfg.get("overwrite", False)
            self.ent_id_sep = cfg.get("ent_id_sep", DEFAULT_ENT_ID_SEP)
        else:
            self.add_patterns(cfg)
        return self

    def to_bytes(self: SpaczzRuler, **kwargs: Any) -> bytes:
        """Serialize the spaczz ruler patterns to a bytestring.

        Args:
            **kwargs: Other config paramters, mostly for consistency.

        Returns:
            The serialized patterns.
        """
        serial = OrderedDict(
            (
                ("overwrite", self.overwrite),
                ("ent_id_sep", self.ent_id_sep),
                ("patterns", self.patterns),
                ("defaults", self.defaults),
            )
        )
        return srsly.msgpack_dumps(serial)

    def from_disk(
        self: SpaczzRuler, path: Union[str, Path], **kwargs: Any
    ) -> SpaczzRuler:
        """Load the spaczz ruler from a file.

        Expects a file containing newline-delimited JSON (JSONL)
        with one entry per line.

        Args:
            path: The JSONL file to load.
            **kwargs: Other config paramters, mostly for consistency.

        Returns:
            The loaded spaczz ruler.
        """
        path = ensure_path(path)
        depr_patterns_path = path.with_suffix(".jsonl")
        if depr_patterns_path.is_file():
            patterns = srsly.read_jsonl(depr_patterns_path)
            self.add_patterns(patterns)
        else:
            cfg = {}
            deserializers_patterns = {
                "patterns": lambda p: self.add_patterns(
                    srsly.read_jsonl(p.with_suffix(".jsonl"))
                )
            }
            deserializers_cfg = {"cfg": lambda p: cfg.update(srsly.read_json(p))}
            read_from_disk(path, deserializers_cfg, {})
            self.overwrite = cfg.get("overwrite", False)
            self.defaults = cfg.get("defaults", {})
            if self.defaults.get("fuzzy_defaults"):
                self.fuzzy_matcher = FuzzyMatcher(
                    self.nlp.vocab, **self.defaults["fuzzy_defaults"]
                )
            if self.defaults.get("regex_defaults"):
                self.regex_matcher = RegexMatcher(
                    self.nlp.vocab, **self.defaults["regex_defaults"]
                )
            if self.defaults.get("token_defaults"):
                self.token_matcher = TokenMatcher(
                    self.nlp.vocab, **self.defaults["token_defaults"]
                )
            self.ent_id_sep = cfg.get("ent_id_sep", DEFAULT_ENT_ID_SEP)
            read_from_disk(path, deserializers_patterns, {})
        return self

    def to_disk(self: SpaczzRuler, path: Union[str, Path], **kwargs: Any) -> None:
        """Save the spaczz ruler patterns to a directory.

        The patterns will be saved as newline-delimited JSON (JSONL).

        Args:
            path: The JSONL file to save.
            **kwargs: Other config paramters, mostly for consistency.
        """
        path = ensure_path(path)
        cfg = {
            "overwrite": self.overwrite,
            "defaults": self.defaults,
            "ent_id_sep": self.ent_id_sep,
        }
        serializers = {
            "patterns": lambda p: srsly.write_jsonl(
                p.with_suffix(".jsonl"), self.patterns
            ),
            "cfg": lambda p: srsly.write_json(p, cfg),
        }
        if path.suffix == ".jsonl":  # user wants to save only JSONL
            srsly.write_jsonl(path, self.patterns)
        else:
            write_to_disk(path, serializers, {})

    def _add_patterns(
        self: SpaczzRuler,
        fuzzy_patterns: List[Dict[str, Any]],
        regex_patterns: List[Dict[str, Any]],
        token_patterns: List[Dict[str, Any]],
    ) -> None:
        """Helper function for add_patterns."""
        for entry in fuzzy_patterns + regex_patterns + token_patterns:
            label = entry["label"]
            if "id" in entry:
                ent_label = label
                label = self._create_label(label, entry["id"])
                self._ent_ids[label] = (ent_label, entry["id"])
            pattern = entry["pattern"]
            if isinstance(pattern, Doc):
                self.fuzzy_patterns[label]["patterns"].append(pattern)
                self.fuzzy_patterns[label]["kwargs"].append(entry["kwargs"])
            elif isinstance(pattern, str):
                self.regex_patterns[label]["patterns"].append(pattern)
                self.regex_patterns[label]["kwargs"].append(entry["kwargs"])
            elif isinstance(pattern, list):
                self.token_patterns[label].append(pattern)
            else:
                raise ValueError(
                    (
                        "One or more patterns do not conform",
                        "to spaczz pattern structure:",
                        "{label (str), pattern (str or list), type (str),",
                        "optional kwargs (dict[str, Any]),",
                        "and optional id (str)}.",
                    )
                )
        for label, patterns in self.fuzzy_patterns.items():
            self.fuzzy_matcher.add(label, patterns["patterns"], patterns["kwargs"])
        for label, patterns in self.regex_patterns.items():
            self.regex_matcher.add(label, patterns["patterns"], patterns["kwargs"])
        for label, _token_patterns in self.token_patterns.items():
            self.token_matcher.add(label, _token_patterns)

    def _create_label(self: SpaczzRuler, label: str, ent_id: Union[str, None]) -> str:
        """Join Entity label with ent_id if the pattern has an id attribute.

        Args:
            label: The entity label.
            ent_id: The optional entity id.

        Returns:
            The label and ent_id joined with configured ent_id_sep.
        """
        if isinstance(ent_id, str):
            label = f"{label}{self.ent_id_sep}{ent_id}"
        return label

    def _split_label(self: SpaczzRuler, label: str) -> Tuple[str, Union[str, None]]:
        """Split Entity label into ent_label and ent_id if it contains self.ent_id_sep.

        Args:
            label: The value of label in a pattern entry

        Returns:
            The separated ent_label and optional ent_id.
        """
        if self.ent_id_sep in label:
            ent_label, ent_id = label.rsplit(self.ent_id_sep, 1)
            return ent_label, ent_id
        else:
            ent_label = label
            return ent_label, None

    @staticmethod
    def _filter_overlapping_matches(
        matches: List[Tuple[str, int, int]],
        lookup: DefaultDict[str, Dict[Tuple[str, int, int], Any]],
    ) -> Tuple[
        List[Tuple[str, int, int]], DefaultDict[str, Dict[Tuple[str, int, int], Any]]
    ]:
        """Prevents multiple match spans from overlapping.

        Expects matches to be pre-sorted by matcher priority,
        with each matcher's matches being pre-sorted by descending length,
        then ascending start index, then descending match score
        If more than one match span includes the same tokens
        the first of these match spans in matches is kept.

        It also removes non-kept matches from the lookup dict as well.

        Args:
            matches: List of match span tuples
                (match_id, start_index, end_index).
            lookup: Match ratio, count and detail values in
                a `defaultdict(dict)`.

        Returns:
            The filtered list of match span tuples and the lookup dict.
        """
        filtered_matches: List[Tuple[str, int, int]] = []
        for match in matches:
            if not set(range(match[1], match[2])).intersection(
                chain(*[set(range(n[1], n[2])) for n in filtered_matches])
            ):
                filtered_matches.append(match)
                if match in lookup["ratios"]:
                    _ = lookup["counts"].pop(match, None)
                    _ = lookup["details"].pop(match, None)
                elif match in lookup["counts"]:
                    _ = lookup["details"].pop(match, None)
        return filtered_matches, lookup

    @staticmethod
    def _update_custom_attrs(
        span: Span,
        match_id: str,
        lookup: DefaultDict[str, Dict[Tuple[str, int, int], Any]],
    ) -> Span:
        """Update custom attributes for matches."""
        ratio = lookup["ratios"].get((match_id, span.start, span.end))
        counts = lookup["counts"].get((match_id, span.start, span.end))
        details = lookup["details"].get((match_id, span.start, span.end))
        for token in span:
            token._.spaczz_token = True
            if ratio:
                token._.spaczz_ratio = ratio
                token._.spaczz_type = "fuzzy"
            elif counts:
                token._.spaczz_counts = counts
                token._.spaczz_type = "regex"
            elif details:
                token._.spaczz_details = details
                token._.spaczz_type = "token"
        return span



"""Module for SimilaritySearcher: vector similarity matching in spaCy `Doc` objects."""
from __future__ import annotations

from typing import Any, Union
import warnings

from spacy.tokens import Doc, Span, Token
from spacy.vocab import Vocab

from . import _PhraseSearcher
from ..exceptions import MissingVectorsWarning


class SimilaritySearcher(_PhraseSearcher):
    """Class for similarity matching in spacy `Doc` objects.

    Similarity matching is done on the token level.
    The class provides methods for finding the best similarity match
    span in a `Doc`, the n best similarity matched spans in a `Doc`,
    and similarity matching between any two given spaCy containers
    (`Doc`, `Span`, `Token`).

    Similarity matching uses spaCy word vectors if available,
    therefore spaCy vocabs without word vectors may not produce
    useful results. The spaCy medium and large English models provide
    work vectors that will work for this purpose.

    Searching in/with spaCy Docs that do not have vector values
    will always return a similarity score of 0.

    Warnings from spaCy about the above two scenarios are suppressed
    for convenience. However, spaczz will still warn about the former.

    Attributes:
        vocab (Vocab): The shared vocabulary.
            Included for consistency and potential future-state.
    """

    def __init__(self: SimilaritySearcher, vocab: Vocab) -> None:
        """Initializes a similarity searcher.

        Args:
            vocab: A spaCy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are mostly pure-Python
                currently and do not share vocabulary
                with spaCy pipelines.

        Warnings:
            MissingVectorsWarning:
                If vocab does not contain any word vectors.
        """
        super().__init__(vocab=vocab)
        if vocab.vectors.n_keys == 0:
            warnings.warn(
                """The spaCy Vocab object has no word vectors.
                Similarity results may not be useful.""",
                MissingVectorsWarning,
            )

    def compare(
        self: SimilaritySearcher,
        query: Union[Doc, Span, Token],
        reference: Union[Doc, Span, Token],
        *args: Any,
        **kwargs: Any
    ) -> int:
        """Peforms similarity matching between two spaCy container objects.

        spaCy containers are `Doc`, `Span`, or `Token` objects.

        Args:
            query: First container for comparison.
            reference: Second container for comparison.
            *args: Overflow for child positional arguments.
            **kwargs: Overflow for child keyword arguments.


        Returns:
            The similarity score between `query` and `reference` as an `int`.

        Example:
            >>> import spacy
            >>> from spaczz.search import SimilaritySearcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> searcher = SimilaritySearcher(nlp.vocab)
            >>> doc1 = nlp("I like apples.")
            >>> doc2 = nlp("I like grapes.")
            >>> searcher.compare(doc1, doc2)
            94
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            if query.vector_norm and reference.vector_norm:
                return round(query.similarity(reference) * 100)
            else:
                return 0



"""Module for RegexSearcher: multi-token regex matching in spaCy `Doc` objects."""
from __future__ import annotations

from typing import List, Tuple, Union

from spacy.tokens import Doc, Span
from spacy.vocab import Vocab

from ..process import map_chars_to_tokens
from ..regex import RegexConfig


class RegexSearcher:
    """Class for multi-token regex matching in spacy `Doc` objects.

    Regex matching is done on the character level and then
    mapped back to tokens.

    Attributes:
        vocab (Vocab): The shared vocabulary.
            Included for consistency and potential future-state.
        _config (RegexConfig): The regex config used with the
            regex searcher.
    """

    def __init__(
        self: RegexSearcher, vocab: Vocab, config: Union[str, RegexConfig] = "default"
    ) -> None:
        """Initializes the regex searcher with the given config.

        Args:
            vocab: A spaCy `Vocab` object.
                Purely for consistency between spaCy
                and spaczz matcher APIs for now.
                spaczz matchers are mostly pure-Python
                currently and do not share vocabulary
                with spaCy pipelines.
            config: Provides the class with predefind regex patterns.
                Uses the default config if `"default"`, an empty config if `"empty"`,
                or a custom config by passing a `RegexConfig` object.
                Default is `"default"`.

        Raises:
            TypeError: If config is not a `RegexConfig` object.
        """
        self.vocab = vocab
        if config == "default":
            self._config = RegexConfig(empty=False)
        elif config == "empty":
            self._config = RegexConfig(empty=True)
        else:
            if isinstance(config, RegexConfig):
                self._config = config
            else:
                raise TypeError(
                    (
                        "config must be one of the strings 'default' or 'empty',",
                        "or a RegexConfig object not,",
                        f"config of type: {type(config)}.",
                    )
                )

    def match(
        self: RegexSearcher,
        doc: Doc,
        query: str,
        partial: bool = True,
        predef: bool = False,
    ) -> List[Tuple[int, int, Tuple[int, int, int]]]:
        """Returns all the regex matches within a doc.

        Matches on the character level and then maps matches back
        to tokens. If a character cannot be mapped back to a token it means
        it is a space tokens are split on, which happens when regex matches
        produce leading or trailing whitespace. Confirm your regex pattern
        will not do this to avoid this issue.

        To utilize regex flags, use inline flags.

        Args:
            doc: `Doc` object to search over.
            query: A string to be compiled to regex,
                or the key name of a predefined regex pattern.
            partial: Whether partial matches should be extended
                to existing span boundaries in doc or not, i.e.
                the regex only matches part of a token or span.
                Default is `True`.
            predef: Whether regex should be interpreted as a key to
                a predefined regex pattern or not. Default is `False`.
                The included regexes are:
                `"dates"`
                `"times"`
                `"phones"`
                `"phones_with_exts"`
                `"links"`
                `"emails"`
                `"ips"`
                `"ipv6s"`
                `"prices"`
                `"hex_colors"`
                `"credit_cards"`
                `"btc_addresses"`
                `"street_addresses"`
                `"zip_codes"`
                `"po_boxes"`
                `"ssn_number"`.

        Returns:
            A list of span start index, end index, fuzzy change count tuples.

        Raises:
            TypeError: If `doc` is not a `Doc`
            TypeError: If `query` is not a `str`.

        Example:
            >>> import spacy
            >>> from spaczz.search import RegexSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = RegexSearcher(nlp.vocab)
            >>> doc = nlp("My phone number is (555) 555-5555.")
            >>> searcher.match(doc, "phones", predef=True)
            [(4, 10, (0, 0, 0))]
        """
        if not isinstance(doc, Doc):
            raise TypeError("doc must be a Doc object.")
        if isinstance(query, str):
            compiled_regex = self._config.parse_regex(query, predef)
        else:
            raise TypeError(f"query must be a str, not {type(query)}.")
        matches = []
        chars_to_tokens = map_chars_to_tokens(doc)
        for match in compiled_regex.finditer(doc.text):
            start, end = match.span()
            counts = match.fuzzy_counts
            span = doc.char_span(start, end)
            if span:
                matches.append((span, counts))
            else:
                if partial:
                    start_token = chars_to_tokens.get(start)
                    end_token = chars_to_tokens.get(end)
                    if start_token and end_token:
                        span = Span(doc, start_token, end_token + 1)
                        matches.append((span, counts))
        if matches:
            return [(match[0].start, match[0].end, match[1]) for match in matches]
        else:
            return []



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['spaczz', 'spaczz.matcher', 'spaczz.pipeline', 'spaczz.regex', 'spaczz.search']

package_data = \
{'': ['*']}

install_requires = \
['rapidfuzz>=1,<2', 'regex>=2020.6.8', 'spacy>=2.2.0,<4.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.0.0']}

entry_points = \
{'spacy_factories': ['spaczz_ruler = spaczz.pipeline:SpaczzRuler']}

setup_kwargs = {
    'name': 'spaczz',
    'version': '0.5.1',
    'description': 'Adds fuzzy matching and additional regex matching support to spaCy.',
    'long_description': '[![Tests](https://github.com/gandersen101/spaczz/workflows/Tests/badge.svg)](https://github.com/gandersen101/spaczz/actions?workflow=Tests)\n[![Codecov](https://codecov.io/gh/gandersen101/spaczz/branch/master/graph/badge.svg)](https://codecov.io/gh/gandersen101/spaczz)\n[![PyPI](https://img.shields.io/pypi/v/spaczz.svg)](https://pypi.org/project/spaczz/)\n[![Read the Docs](https://readthedocs.org/projects/spaczz/badge/)](https://spaczz.readthedocs.io/)\n\n# spaczz: Fuzzy matching and more for spaCy\n\nSpaczz provides fuzzy matching and additional regex matching functionality for [spaCy](https://spacy.io/).\nSpaczz\'s components have similar APIs to their spaCy counterparts and spaczz pipeline components can integrate into spaCy pipelines where they can be saved/loaded as models.\n\nFuzzy matching is currently performed with matchers from [RapidFuzz](https://github.com/maxbachmann/rapidfuzz)\'s fuzz module and regex matching currently relies on the [regex](https://pypi.org/project/regex/) library. Spaczz certainly takes additional influence from other libraries and resources. For additional details see the references section.\n\n**Supports spaCy v3 and v2 (>= 2.2)!**\n\nSpaczz has been tested on Ubuntu 18.04, MacOS 10.15, and Windows Server 2019.\n\n*v0.5.1 Release Notes:*\n- *Minor updates to allowed dependency versions and CI.*\n- *Switched back to using typing types instead of generic types because spaCy v3 uses Pydantic and Pydantic does not support generic types in Python < 3.9. I don\'t know if this would actually cause any issues but I am playing it safe. Potentially more changes for spaczz to play nicely with Pydantic to follow.*\n\n*v0.5.0 Release Notes:*\n- *Support for spaCy v3.*\n- *If using spaCy v3, the `SpaczzRuler` optional arguments no longer need to be prepended with `"spaczz_"`. This will still work in most cases offering some backwards compatibility. However, optional arguments prepended with `"spaczz_"` will not work with spaCy v3\'s new `spacy.load` and `nlp.add_pipe` config driven APIs. It is therefore recommended that users move away from using the prepended versions if using spaCy v3. It should be noted however that the prepended arguments are still necessary if using spaczz with spaCy v2.*\n- *`Matcher.pipe` methods are now deprecated in accordance with spaCy v3.*\n- *`spaczz_span` custom attribute is deprecated in favor of `spaczz_ent`. They both have the same functionality but the `spaczz_ent` name makes more sense.*\n\nPlease see the [changelog](https://github.com/gandersen101/spaczz/blob/master/CHANGELOG.md) for previous release notes. This will eventually be moved to the [Read the Docs](https://spaczz.readthedocs.io/en/latest/) page.\n\n<h1>Table of Contents<span class="tocSkip"></span></h1>\n<div class="toc"><ul class="toc-item"><li><span><a href="#Installation" data-toc-modified-id="Installation-1">Installation</a></span></li><li><span><a href="#Basic-Usage" data-toc-modified-id="Basic-Usage-2">Basic Usage</a></span><ul class="toc-item"><li><span><a href="#FuzzyMatcher" data-toc-modified-id="FuzzyMatcher-2.1">FuzzyMatcher</a></span></li><li><span><a href="#RegexMatcher" data-toc-modified-id="RegexMatcher-2.2">RegexMatcher</a></span></li><li><span><a href="#SimilarityMatcher" data-toc-modified-id="SimilarityMatcher-2.3">SimilarityMatcher</a></span></li><li><span><a href="#TokenMatcher" data-toc-modified-id="TokenMatcher-2.4">TokenMatcher</a></span></li><li><span><a href="#SpaczzRuler" data-toc-modified-id="SpaczzRuler-2.5">SpaczzRuler</a></span></li><li><span><a href="#Custom-Attributes" data-toc-modified-id="Custom-Attributes-2.6">Custom Attributes</a></span></li><li><span><a href="#Saving/Loading" data-toc-modified-id="Saving/Loading-2.7">Saving/Loading</a></span></li></ul></li><li><span><a href="#Known-Issues" data-toc-modified-id="Known-Issues-3">Known Issues</a></span><ul class="toc-item"><li><span><a href="#Performance" data-toc-modified-id="Performance-3.1">Performance</a></span></li><li><span><a href="#SpaczzRuler-Inconsistencies" data-toc-modified-id="SpaczzRuler-Inconsistencies-3.2">SpaczzRuler Inconsistencies</a></span></li></ul></li><li><span><a href="#Roadmap" data-toc-modified-id="Roadmap-4">Roadmap</a></span></li><li><span><a href="#Development" data-toc-modified-id="Development-5">Development</a></span></li><li><span><a href="#References" data-toc-modified-id="References-6">References</a></span></li></ul></div>\n\n## Installation\n\nSpaczz can be installed using pip.\n\n\n```python\npip install spaczz\n```\n\n## Basic Usage\n\nSpaczz\'s primary features are the `FuzzyMatcher`, `RegexMatcher`, and "fuzzy" `TokenMatcher` that function similarly to spaCy\'s `Matcher` and `PhraseMatcher`, and the `SpaczzRuler` which integrates the spaczz matchers into a spaCy pipeline component similar to spaCy\'s `EntityRuler`.\n\n### FuzzyMatcher\n\nThe basic usage of the fuzzy matcher is similar to spaCy\'s `PhraseMatcher` except it returns the fuzzy ratio along with match id, start and end information, so make sure to include a variable for the ratio when unpacking results.\n\n\n```python\nimport spacy\nfrom spaczz.matcher import FuzzyMatcher\n\nnlp = spacy.blank("en")\ntext = """Grint Anderson created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the US."""  # Spelling errors intentional.\ndoc = nlp(text)\n\nmatcher = FuzzyMatcher(nlp.vocab)\nmatcher.add("NAME", [nlp("Grant Andersen")])\nmatcher.add("GPE", [nlp("Nashville")])\nmatches = matcher(doc)\n\nfor match_id, start, end, ratio in matches:\n    print(match_id, doc[start:end], ratio)\n```\n\n    NAME Grint Anderson 86\n    GPE Nashv1le 82\n\n\nUnlike spaCy matchers, spaczz matchers are written in pure Python. While they are required to have a spaCy vocab passed to them during initialization, this is purely for consistency as the spaczz matchers do not use currently use the spaCy vocab. This is why the `match_id` above is simply a string instead of an integer value like in spaCy matchers.\n\nSpaczz matchers can also make use of on-match rules via callback functions. These on-match callbacks need to accept the matcher itself, the doc the matcher was called on, the match index and the matches produced by the matcher.\n\n\n```python\nimport spacy\nfrom spacy.tokens import Span\nfrom spaczz.matcher import FuzzyMatcher\n\nnlp = spacy.blank("en")\ntext = """Grint Anderson created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the US."""  # Spelling errors intentional.\ndoc = nlp(text)\n\n\ndef add_name_ent(matcher, doc, i, matches):\n    """Callback on match function. Adds "NAME" entities to doc."""\n    # Get the current match and create tuple of entity label, start and end.\n    # Append entity to the doc\'s entity. (Don\'t overwrite doc.ents!)\n    _match_id, start, end, _ratio = matches[i]\n    entity = Span(doc, start, end, label="NAME")\n    doc.ents += (entity,)\n\n\nmatcher = FuzzyMatcher(nlp.vocab)\nmatcher.add("NAME", [nlp("Grant Andersen")], on_match=add_name_ent)\nmatches = matcher(doc)\n\nfor ent in doc.ents:\n    print((ent.text, ent.start, ent.end, ent.label_))\n```\n\n    (\'Grint Anderson\', 0, 2, \'NAME\')\n\n\nLike spaCy\'s `EntityRuler`, a very similar entity updating logic has been implemented in the `SpaczzRuler`. The `SpaczzRuler` also takes care of handling overlapping matches. It is discussed in a later section.\n\nUnlike spaCy\'s matchers, rules added to spaczz matchers have optional keyword arguments that can modify the matching behavior. Take the below fuzzy matching examples:\n\n\n```python\nimport spacy\nfrom spaczz.matcher import FuzzyMatcher\n\nnlp = spacy.blank("en")\n# Let\'s modify the order of the name in the text.\ntext = """Anderson, Grint created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the US."""  # Spelling errors intentional.\ndoc = nlp(text)\n\nmatcher = FuzzyMatcher(nlp.vocab)\nmatcher.add("NAME", [nlp("Grant Andersen")])\nmatches = matcher(doc)\n\n# The default fuzzy matching settings will not find a match.\nfor match_id, start, end, ratio in matches:\n    print(match_id, doc[start:end], ratio)\n```\n\nNext we change the fuzzy matching behavior for the "NAME" rule.\n\n\n```python\nimport spacy\nfrom spaczz.matcher import FuzzyMatcher\n\nnlp = spacy.blank("en")\n# Let\'s modify the order of the name in the text.\ntext = """Anderson, Grint created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the US."""  # Spelling errors intentional.\ndoc = nlp(text)\n\nmatcher = FuzzyMatcher(nlp.vocab)\nmatcher.add("NAME", [nlp("Grant Andersen")], kwargs=[{"fuzzy_func": "token_sort"}])\nmatches = matcher(doc)\n\n# The default fuzzy matching settings will not find a match.\nfor match_id, start, end, ratio in matches:\n    print(match_id, doc[start:end], ratio)\n```\n\n    NAME Anderson, Grint 86\n\n\nThe full list of keyword arguments available for fuzzy matching rules includes:\n\n- `fuzzy_func`: Key name of fuzzy matching function to use. All rapidfuzz matching functions with default settings are available. Default is `"simple"`:\n    - "simple" = `ratio`\n    - "partial" = `partial_ratio`\n    - "token_set" = `token_set_ratio`\n    - "token_sort" = `token_sort_ratio`\n    - "partial_token_set" = `partial_token_set_ratio`\n    - "partial_token_sort" = `partial_token_sort_ratio`\n    - "quick" = `QRatio`\n    - "weighted" = `WRatio`\n    - "token" = `token_ratio`,\n    - "partial_token" = `partial_token_ratio`\n       Default is `"simple"`.\n- `ignore_case`: If strings should be lower-cased before comparison or not. Default is `True`.\n- `flex`: Number of tokens to move match match boundaries left and right during optimization. Can be an integer value with a max of `len(query)` and a min of `0` (will warn and change if higher or lower),or the strings "max", "min", or "default". Default is `"default"`: `len(query) // 2`.\n- `min_r1`: Minimum match ratio required forselection during the intial search over doc. If `flex == 0`, `min_r1` will be overwritten by `min_r2`. If `flex > 0`, `min_r1` must be lower than `min_r2` and "low" in general because match boundaries are not flexed initially. Default is `50`.\n- `min_r2`: Minimum match ratio required for selection during match optimization. Needs to be higher than `min_r1` and "high" in general to ensure only quality matches are returned. Default is `75`.\n- `thresh`: If this ratio is exceeded in initial scan, and `flex > 0`, no optimization will be attempted. If `flex == 0`, `thresh` has no effect. Default is `100`.\n\n### RegexMatcher\n\nThe basic usage of the regex matcher is also fairly similar to spaCy\'s `PhraseMatcher`. It accepts regex patterns as strings so flags must be inline. Regexes are compiled with the [regex](https://pypi.org/project/regex/) package so approximate "fuzzy" matching is supported. To provide access to these "fuzzy" match results the matcher returns the fuzzy count values along with match id, start and end information, so make sure to include a variable for the counts when unpacking results.\n\n\n```python\nimport spacy\nfrom spaczz.matcher import RegexMatcher\n\nnlp = spacy.blank("en")\ntext = """Anderson, Grint created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the US."""  # Spelling errors intentional.\ndoc = nlp(text)\n\nmatcher = RegexMatcher(nlp.vocab)\n# Use inline flags for regex strings as needed\nmatcher.add(\n    "APT",\n    [\n        r"""(?ix)((?:apartment|apt|building|bldg|floor|fl|suite|ste|unit\n|room|rm|department|dept|row|rw)\\.?\\s?)#?\\d{1,4}[a-z]?"""\n    ],\n)  # Not the most robust regex.\nmatcher.add("GPE", [r"(USA){d<=1}"])  # Fuzzy regex.\nmatches = matcher(doc)\n\nfor match_id, start, end, counts in matches:\n    print(match_id, doc[start:end], counts)\n```\n\n    APT Apt 5 (0, 0, 0)\n    GPE US (0, 0, 1)\n\n\nSpaczz matchers can also make use of on-match rules via callback functions. These on-match callbacks need to accept the matcher itself, the doc the matcher was called on, the match index and the matches produced by the matcher. See the fuzzy matcher usage example above for details.\n\nLike the fuzzy matcher, the regex matcher has optional keyword arguments that can modify matching behavior. Take the below regex matching example.\n\n\n```python\nimport spacy\nfrom spaczz.matcher import RegexMatcher\n\nnlp = spacy.blank("en")\ntext = """Anderson, Grint created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the USA."""  # Spelling errors intentional. Notice \'USA\' here.\ndoc = nlp(text)\n\nmatcher = RegexMatcher(nlp.vocab)\n# Use inline flags for regex strings as needed\nmatcher.add(\n    "STREET", ["street_addresses"], kwargs=[{"predef": True}]\n)  # Use predefined regex by key name.\n# Below will not expand partial matches to span boundaries.\nmatcher.add("GPE", [r"(?i)[U](nited|\\.?) ?[S](tates|\\.?)"], kwargs=[{"partial": False}])\nmatches = matcher(doc)\n\nfor match_id, start, end, counts in matches:\n    print(\n        match_id, doc[start:end], counts\n    )  # comma in result isn\'t ideal - see "Roadmap"\n```\n\n    STREET 555 Fake St, (0, 0, 0)\n\n\nThe full list of keyword arguments available for regex matching rules includes:\n\n- `partial`: Whether partial matches should be extended to existing span boundaries in doc or not, i.e. the regex only matches part of a token or span. Default is True.\n- `predef`: Whether the regex string should be interpreted as a key to a predefined regex pattern or not. Default is False. The included regexes are:\n    - `"dates"`\n    - `"times"`\n    - `"phones"`\n    - `"phones_with_exts"`\n    - `"links"`\n    - `"emails"`\n    - `"ips"`\n    - `"ipv6s"`\n    - `"prices"`\n    - `"hex_colors"`\n    - `"credit_cards"`\n    - `"btc_addresses"`\n    - `"street_addresses"`\n    - `"zip_codes"`\n    - `"po_boxes"`\n    - `"ssn_number"`\n\nThe above patterns are the same that the [commonregex](https://github.com/madisonmay/CommonRegex) package provides.\n\n### SimilarityMatcher\n\nThe basic usage of the similarity matcher is similar to spaCy\'s `PhraseMatcher` except it returns the vector similarity ratio along with match id, start and end information, so make sure to include a variable for the ratio when unpacking results.\n\nIn order to produce meaningful results from the similarity matcher, a spaCy model with word vectors (ex. medium or large English models) must be used to initialize the matcher, process the target document, and process any patterns added.\n\n\n```python\nimport spacy\nfrom spaczz.matcher import SimilarityMatcher\n\nnlp = spacy.load("en_core_web_md")\ntext = "I like apples, grapes and bananas."\ndoc = nlp(text)\n\n# lowering min_r2 from default of 75 to produce matches in this example\nmatcher = SimilarityMatcher(nlp.vocab, min_r2=65)\nmatcher.add("FRUIT", [nlp("fruit")])\nmatches = matcher(doc)\n\nfor match_id, start, end, ratio in matches:\n    print(match_id, doc[start:end], ratio)\n```\n\n    FRUIT apples 72\n    FRUIT grapes 72\n    FRUIT bananas 68\n\n\nPlease note that even for the mostly pure-Python spaczz, this process is currently extremely slow so be mindful of the scope in which it is applied. Enabling GPU support in spaCy ([see here](https://spacy.io/usage#gpu)) should improve the speed somewhat, but I believe the process will still be bottlenecked in the pure-Python search algorithm until I develop a better search algorithm and/or drop the search to lower-level code (ex C).\n\nAlso as a somewhat experimental feature, the similarity matcher is not currently part of the `SpaczzRuler` nor does it have a separate ruler. If you need to add similarity matches to a doc\'s entities you will need to use an on-match callback for the time being. Please see the fuzzy matcher on-match callback example above for ideas. If there is enough interest in integrating/creating a ruler for the similarity matcher this can be done.\n\nThe full list of keyword arguments available for similarity matching rules includes:\n\n- `flex`: Number of tokens to move match span boundaries left and right during match optimization. Can be an integer value with a max of `len(query)` and a min of `0` (will warn and change if higher or lower), `"max"`, `"min"`, or `"default"`. Default is `"default"`: `len(query) // 2`.\n- `min_r1`: Minimum similarity match ratio required for selection during the intial search over doc. This should be lower than `min_r2` and "low" in general because match span boundaries are not flexed initially. `0` means all spans of query length in doc will have their boundaries flexed and will be re-compared during match optimization. Lower `min_r1` will result in more fine-grained matching but will run slower. Default is `50`.\n- `min_r2`: Minimum similarity match ratio required for selection during match optimization. Should be higher than `min_r1` and "high" in general to ensure only quality matches are returned. Default is `75`.\n- `thresh`: If this ratio is exceeded in initial scan no optimization will be attempted. Default is `100`.\n\n### TokenMatcher\n\nThe basic usage of the token matcher is similar to spaCy\'s `Matcher`. It accepts labeled patterns in the form of lists of dictionaries where each list describes an individual pattern and each dictionary describes an individual token.\n\nThe token matcher accepts all the same token attributes and pattern syntax as it\'s spaCy counterpart but adds fuzzy and fuzzy-regex support.\n\n`"FUZZY"` and `"FREGEX"` are the two additional spaCy token pattern options.\n\nFor example:\n    `{"TEXT": {"FREGEX": "(database){e<=1}"}},`\n    `{"LOWER": {"FUZZY": "access", "MIN_R": 85, "FUZZY_FUNC": "quick_lev"}}`\n\n**Make sure to use uppercase dictionary keys in patterns.**\n\n\n```python\nimport spacy\nfrom spaczz.matcher import TokenMatcher\n\n# Using model results like POS tagging in token patterns requires model that provides these.\nnlp = spacy.load("en_core_web_md")\ntext = """The manager gave me SQL databesE acess so now I can acces the Sequal DB.\nMy manager\'s name is Grfield"""\ndoc = nlp(text)\n\nmatcher = TokenMatcher(vocab=nlp.vocab)\nmatcher.add(\n    "DATA",\n    [\n        [\n            {"TEXT": "SQL"},\n            {"LOWER": {"FREGEX": "(database){s<=1}"}},\n            {"LOWER": {"FUZZY": "access"}},\n        ],\n        [{"TEXT": {"FUZZY": "Sequel"}, "POS": "PROPN"}, {"LOWER": "db"}],\n    ],\n)\nmatcher.add("NAME", [[{"TEXT": {"FUZZY": "Garfield"}}]])\nmatches = matcher(doc)\n\nfor match_id, start, end, _ in matches:  # Note the _ here. Explained below.\n    print(match_id, doc[start:end])\n```\n\n    DATA SQL databesE acess\n    DATA Sequal DB\n    NAME Grfield\n\n\nPlease note that the way the token matcher is implemented does not currently have a way to return fuzzy ratio or fuzzy-regex counts like the fuzzy matcher and regex matcher provide. To keep the API consistent, the token matcher returns a placeholder of `None` as the fourth element of the tuples it returns, so be sure to account for this like we did with `_` in unpacking above.\n\nAlso, even though the token matcher can be a drop-in replacement for spaCy\'s `Matcher`, it is still recommended to use spaCy\'s `Matcher` if you do not need the spaczz token matcher\'s fuzzy capabilities - it will slow processing down unnecessarily.\n\n### SpaczzRuler\n\nThe spaczz ruler combines the fuzzy and regex phrase matchers, and the "fuzzy" token matcher, into one pipeline component that can update a doc entities similar to spaCy\'s `EntityRuler`.\n\nPatterns must be added as an iterable of dictionaries in the format of *{label (str), pattern(str or list), type(str), optional kwargs (dict), and optional id (str)}*.\n\nFor example, a fuzzy phrase pattern:\n\n`{\'label\': \'ORG\', \'pattern\': \'Apple\', \'type\': \'fuzzy\', \'kwargs\': {\'min_r2\': 90}}`\n\nOr, a token pattern:\n\n`{\'label\': \'ORG\', \'pattern\': [{\'TEXT\': {\'FUZZY\': \'Apple\'}}], \'type\': \'token\'}`\n\n\n```python\nimport spacy\nfrom spaczz.pipeline import SpaczzRuler\n\nnlp = spacy.blank("en")\ntext = """Anderson, Grint created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the USA.\nSome of his favorite bands are Converg and Protet the Zero."""  # Spelling errors intentional.\ndoc = nlp(text)\n\npatterns = [\n    {\n        "label": "NAME",\n        "pattern": "Grant Andersen",\n        "type": "fuzzy",\n        "kwargs": {"fuzzy_func": "token_sort"},\n    },\n    {\n        "label": "STREET",\n        "pattern": "street_addresses",\n        "type": "regex",\n        "kwargs": {"predef": True},\n    },\n    {"label": "GPE", "pattern": "Nashville", "type": "fuzzy"},\n    {\n        "label": "ZIP",\n        "pattern": r"\\b(?:55554){s<=1}(?:(?:[-\\s])?\\d{4}\\b)",\n        "type": "regex",\n    },  # fuzzy regex\n    {"label": "GPE", "pattern": "(?i)[U](nited|\\.?) ?[S](tates|\\.?)", "type": "regex"},\n    {\n        "label": "BAND",\n        "pattern": [{"LOWER": {"FREGEX": "(converge){e<=1}"}}],\n        "type": "token",\n    },\n    {\n        "label": "BAND",\n        "pattern": [\n            {"TEXT": {"FUZZY": "Protest"}},\n            {"IS_STOP": True},\n            {"TEXT": {"FUZZY": "Hero"}},\n        ],\n        "type": "token",\n    },\n]\n\nruler = SpaczzRuler(nlp)\nruler.add_patterns(patterns)\ndoc = ruler(doc)\n\nprint("Fuzzy Matches:")\nfor ent in doc.ents:\n    if ent._.spaczz_type == "fuzzy":\n        print((ent.text, ent.start, ent.end, ent.label_, ent._.spaczz_ratio))\n\nprint("\\n", "Regex Matches:", sep="")\nfor ent in doc.ents:\n    if ent._.spaczz_type == "regex":\n        print((ent.text, ent.start, ent.end, ent.label_, ent._.spaczz_counts))\n\nprint("\\n", "Token Matches:", sep="")\nfor ent in doc.ents:\n    if ent._.spaczz_type == "token":\n        # ._.spaczz_details is currently just placeholder value of 1\n        print((ent.text, ent.start, ent.end, ent.label_, ent._.spaczz_details))\n```\n\n    Fuzzy Matches:\n    (\'Anderson, Grint\', 0, 3, \'NAME\', 86)\n    (\'Nashv1le\', 17, 18, \'GPE\', 82)\n\n    Regex Matches:\n    (\'555 Fake St,\', 9, 13, \'STREET\', (0, 0, 0))\n    (\'55555-1234\', 20, 23, \'ZIP\', (1, 0, 0))\n    (\'USA\', 25, 26, \'GPE\', (0, 0, 0))\n\n    Token Matches:\n    (\'Converg\', 34, 35, \'BAND\', 1)\n    (\'Protet the Zero\', 36, 39, \'BAND\', 1)\n\n\nWe see in the example above that we are referencing some custom attributes, which are explained below.\n\nFor more `SpaczzRuler` examples see [here](https://github.com/gandersen101/spaczz/blob/master/examples/fuzzy_matching_tweaks.md). In particular this provides details about the ruler\'s sorting process and fuzzy matching parameters.\n\n### Custom Attributes\n\nSpaczz initializes some custom attributes upon importing. These are under spaCy\'s `._.` attribute and are further prepended with `spaczz_` so there should be not conflicts with your own custom attributes. If there are spaczz will force overwrite them.\n\nThese custom attributes are only set via the spaczz ruler at the token level. Span and doc versions of these attributes are getters that reference the token level attributes.\n\nThe following `Token` attributes are available. All are mutable:\n\n- `spaczz_token`: default = `False`. Boolean that denotes if the token is part of an ent set by the spaczz ruler.\n- `spaczz_type`: default = `None`. String that shows which matcher produced an ent using the token.\n- `spaczz_ratio`: default = `None`. If the token is part of fuzzy-phrase-matched ent, will return fuzzy ratio.\n- `spaczz_counts`: default = `None`. If the token is part of regex-phrase-matched ent, will return fuzzy counts.\n- `spaczz_details`: default = `None`. Placeholder for token matcher fuzzy ratio/counts. To be developed. Will return 1 if the token is part of a "fuzzy"-token-matched ent.\n\nThe following `Span` attributes reference the token attributes included in the span. All are immutable:\n\n- `spaczz_ent`: default = `False`. Boolean that denotes if all tokens in span are part of an ent set by the spaczz ruler.\n- `spaczz_type`: default = `None`. String that denotes which matcher produced an ent using the included tokens.\n- `spaczz_types`: default = `set()`. Set that shows which matchers produced ents using the included tokens. An entity span should only have one type, but this allows you to see the types included in any arbitrary span.\n- `spaczz_ratio`: default = `None`. If all the tokens in span are part of fuzzy-phrase-matched ent, will return fuzzy ratio.\n- `spaczz_counts`: default = `None`. If all the tokens in span are part of regex-phrase-matched ent, will return fuzzy counts.\n- `spaczz_details`: default = `None`. Placeholder for token matcher fuzzy ratio/counts. To be developed. Will return 1 if all the tokens in span are part of a "fuzzy"-token-matched ent.\n\nThe following `Doc` attributes reference the token attributes included in the doc. All are immutable:\n\n- `spaczz_doc`: default = `False`. Boolean that denotes if any tokens in the doc are part of an ent set by the spaczz ruler.\n- `spaczz_types`: default = `set()`. Set that shows which matchers produced ents in the doc.\n\n### Saving/Loading\n\nThe `SpaczzRuler` has it\'s own to/from disk/bytes methods and will accept `config` parameters passed to `spacy.load()`. It also has it\'s own spaCy factory entry point so spaCy is aware of the `SpaczzRuler`. Below is an example of saving and loading a spacy pipeline with the small English model, the `EntityRuler`, and the `SpaczzRuler`.\n\n\n```python\nimport spacy\nfrom spaczz.pipeline import SpaczzRuler\n\nnlp = spacy.load("en_core_web_sm")\ntext = """Anderson, Grint created spaczz in his home at 555 Fake St,\nApt 5 in Nashv1le, TN 55555-1234 in the USA.\nSome of his favorite bands are Converg and Protet the Zero."""  # Spelling errors intentional.\ndoc = nlp(text)\n\nfor ent in doc.ents:\n    print((ent.text, ent.start, ent.end, ent.label_))\n```\n\n    (\'Anderson\', 0, 1, \'PERSON\')\n    (\'Grint\', 2, 3, \'PERSON\')\n    (\'555\', 9, 10, \'CARDINAL\')\n    (\'5\', 15, 16, \'CARDINAL\')\n    (\'TN 55555-1234\', 19, 23, \'DATE\')\n    (\'USA\', 25, 26, \'GPE\')\n    (\'Converg\', 34, 35, \'PERSON\')\n    (\'Protet\', 36, 37, \'PERSON\')\n\n\nWhile spaCy does a decent job of identifying that named entities are present in this example, we can definitely improve the matches - particularly with the types of labels applied.\n\nLet\'s add an entity ruler for some rules-based matches.\n\n\n```python\nfrom spacy.pipeline import EntityRuler\n\nentity_ruler = nlp.add_pipe("entity_ruler", before="ner") #spaCy v3 syntax\nentity_ruler.add_patterns(\n    [{"label": "GPE", "pattern": "Nashville"}, {"label": "GPE", "pattern": "TN"}]\n)\n\ndoc = nlp(text)\n\nfor ent in doc.ents:\n    print((ent.text, ent.start, ent.end, ent.label_))\n```\n\n    (\'Anderson\', 0, 1, \'PERSON\')\n    (\'Grint\', 2, 3, \'PERSON\')\n    (\'555\', 9, 10, \'CARDINAL\')\n    (\'5\', 15, 16, \'CARDINAL\')\n    (\'TN\', 19, 20, \'GPE\')\n    (\'USA\', 25, 26, \'GPE\')\n    (\'Converg\', 34, 35, \'PERSON\')\n    (\'Protet\', 36, 37, \'PERSON\')\n\n\nWe\'re making progress, but Nashville is spelled wrong in the text so the entity ruler does not find it, and we still have other entities to fix/find.\n\nLet\'s add a spaczz ruler to round this pipeline out. We will also include the `spaczz_span` custom attribute in the results to denote which entities were set via spaczz.\n\n\n```python\nspaczz_ruler = nlp.add_pipe("spaczz_ruler", before="ner") #spaCy v3 syntax\nspaczz_ruler.add_patterns(\n    [\n        {\n            "label": "NAME",\n            "pattern": "Grant Andersen",\n            "type": "fuzzy",\n            "kwargs": {"fuzzy_func": "token_sort"},\n        },\n        {\n            "label": "STREET",\n            "pattern": "street_addresses",\n            "type": "regex",\n            "kwargs": {"predef": True},\n        },\n        {"label": "GPE", "pattern": "Nashville", "type": "fuzzy"},\n        {\n            "label": "ZIP",\n            "pattern": r"\\b(?:55554){s<=1}(?:[-\\s]\\d{4})?\\b",\n            "type": "regex",\n        },  # fuzzy regex\n        {\n            "label": "BAND",\n            "pattern": [{"LOWER": {"FREGEX": "(converge){e<=1}"}}],\n            "type": "token",\n        },\n        {\n            "label": "BAND",\n            "pattern": [\n                {"TEXT": {"FUZZY": "Protest"}},\n                {"IS_STOP": True},\n                {"TEXT": {"FUZZY": "Hero"}},\n            ],\n            "type": "token",\n        },\n    ]\n)\n\ndoc = nlp(text)\n\nfor ent in doc.ents:\n    print((ent.text, ent.start, ent.end, ent.label_, ent._.spaczz_ent))\n```\n\n    (\'Anderson, Grint\', 0, 3, \'NAME\', True)\n    (\'555 Fake St,\', 9, 13, \'STREET\', True)\n    (\'5\', 15, 16, \'CARDINAL\', False)\n    (\'Nashv1le\', 17, 18, \'GPE\', True)\n    (\'TN\', 19, 20, \'GPE\', False)\n    (\'55555-1234\', 20, 23, \'ZIP\', True)\n    (\'USA\', 25, 26, \'GPE\', False)\n    (\'Converg\', 34, 35, \'BAND\', True)\n    (\'Protet the Zero\', 36, 39, \'BAND\', True)\n\n\nAwesome! The small English model still makes a named entity recognition mistake ("5" in "Apt 5" as `CARDINAL`), but we\'re satisfied overall.\n\nLet\'s save this pipeline to disk and make sure we can load it back correctly.\n\n\n```python\nnlp.to_disk("./example")\nnlp = spacy.load("./example")\nnlp.pipe_names\n```\n\n\n\n\n    [\'tok2vec\',\n     \'tagger\',\n     \'parser\',\n     \'entity_ruler\',\n     \'spaczz_ruler\',\n     \'ner\',\n     \'attribute_ruler\',\n     \'lemmatizer\']\n\n\n\nWe can even ensure all the spaczz ruler patterns are still present.\n\n\n```python\nspaczz_ruler = nlp.get_pipe("spaczz_ruler")\nspaczz_ruler.patterns\n```\n\n\n\n\n    [{\'label\': \'NAME\',\n      \'pattern\': \'Grant Andersen\',\n      \'type\': \'fuzzy\',\n      \'kwargs\': {\'fuzzy_func\': \'token_sort\'}},\n     {\'label\': \'GPE\', \'pattern\': \'Nashville\', \'type\': \'fuzzy\'},\n     {\'label\': \'STREET\',\n      \'pattern\': \'street_addresses\',\n      \'type\': \'regex\',\n      \'kwargs\': {\'predef\': True}},\n     {\'label\': \'ZIP\',\n      \'pattern\': \'\\\\b(?:55554){s<=1}(?:[-\\\\s]\\\\d{4})?\\\\b\',\n      \'type\': \'regex\'},\n     {\'label\': \'BAND\',\n      \'pattern\': [{\'LOWER\': {\'FREGEX\': \'(converge){e<=1}\'}}],\n      \'type\': \'token\'},\n     {\'label\': \'BAND\',\n      \'pattern\': [{\'TEXT\': {\'FUZZY\': \'Protest\'}},\n       {\'IS_STOP\': True},\n       {\'TEXT\': {\'FUZZY\': \'Hero\'}}],\n      \'type\': \'token\'}]\n\n\n\n## Known Issues\n\n### Performance\n\nThe main reason for spaczz\'s slower speed is that the *c* in it\'s name is not capitalized like it is in spa*C*y.\nSpaczz is written in pure Python and it\'s matchers do not currently utilize spaCy language vocabularies, which means following it\'s logic should be easy to those familiar with Python. However this means spaczz components will run slower and likely consume more memory than their spaCy counterparts, especially as more patterns are added and documents get longer. It is therefore recommended to use spaCy components like the EntityRuler for entities with little uncertainty, like consistent spelling errors. Use spaczz components when there are not viable spaCy alternatives.\n\nI am actively working on performance optimizations to spaczz but it is a gradual process. Algorithmic and optimization suggestions are welcome. I am working on learning C but currently C-based work is outside of my skill set.\n\nThe `FuzzyMatcher`, and even more so, the `SimilarityMatcher` are the slowest spaczz components (although allowing for enough "fuzzy" matches in the `RegexMatcher` can get really slow as well). The primary methods for speeding these components up are by decreasing the `flex` parameter towards `0`, or if `flex > 0`, increasing the `min_r1` parameter towards the value of `min_r2` and/or lowering the `thresh` parameter towards `min_r2`. Be aware that all of these "speed-ups" come at the opportunity cost of potentially improved matches.\n\nAs mentioned in the `SimilarityMatcher` description, utilizing a GPU will also help speed up it\'s matching process.\n\nI will likely try to develop some automated and/or heuristic-based API options (while retaining all the current options) in the future to simplify this "tuning" process.\n\n### SpaczzRuler Inconsistencies\n\nThis one is particularly annoying for me because I built myself into this hole trying to support too much too fast. That being said I have addressed much of this as of spaczz 0.4.2 and will continue to improve these issues.\n\nSpaczz, like spaCy, has undefined behavior for multiple labels (or label/ent_id combos) sharing the same pattern. For example, if you add the pattern `"Ireland"` as both `"GPE"` and `"NAME"` the resulting label is unpredictable. For the most part this isn\'t an issue but spaczz also has to deal with the additional wrinkle of fuzzy matches.\n\nFor example if we are looking for the string `"Ireland"` and have the patterns `["Ireland", "Iceland"]`. Even with a required match ratio of `85` these will both match at `100` and `86` respectively. When just dealing with fuzzy matches this isn\'t an issue as we can sort by descending match ratio. However what if the `"Iceland"` pattern was a regex pattern and it returned a tuple of fuzzy regex counts? Or what if the `"Iceland"` pattern was a token pattern and the `TokenMatcher` does not even currently provide match details?!\n\nThe above problem is twofold. First and foremost, I need to develop a way or ways to compare apples to oranges - fuzzy ratios and fuzzy regex counts. Then I need to figure out how to include match details from the `TokenMatcher` which supports both fuzzy and "fuzzy" regex matches.\n\nFor a short-term solution I am having the entity ruler first go through sorted fuzzy matches, then sorted regex matches, and lastly token matches. Token matches will only be sorted by length of match, not quality, so they may provide inconsistent results. Try to be mindful of your token patterns.\n\nThere is additional logic in place to filter overlapping matches preserving earlier matches over later ones. This order of priority (fuzzy, regex, token) may not be ideal for everyone but adding a way to change the order (say regex patterns first) would a temporary solution to a temporary problem.\n\nPlease bear with me through these growing pains.\n\n## Roadmap\n\nI am always open and receptive to feature requests but just be aware, as a solo-dev with a lot left to learn, development can move pretty slow. The following is my roadmap for spaczz so you can see where issues raised might fit into my current priorities.\n\n**High Priority**\n\n1. Bug fixes - both breaking and behavioral. Hopefully these will be minimal.\n1. Ease of use and error/warning handling and messaging enhancements.\n1. Building out Read the Docs.\n1. A method for comparing fuzzy ratios and fuzzy regex counts.\n1. A way to return match details from the `TokenMatcher`.\n1. Option to prioritize match quality over length and/or weighing options.\n1. Profiling - hopefully to find "easy" performance optimizations.\n\n**Enhancements**\n\n1. API support for adding user-defined regexes to the predefined regex.\n    1. Saving these additional predefined regexes as part of the SpaczzRuler will also be supported.\n1. Entity start/end trimming on the token level to prevent fuzzy and regex phrase matches from starting/ending with unwanted tokens, i.e. spaces/punctuation.\n\n**Long-Horizon Performance Enhancements**\n\n1. Having spaczz matchers utilize spaCy vocabularies.\n1. Rewrite the phrase and token searching algorithms in Cython to utilize C speed.\n    1. Try to integrate closely with spaCy.\n\n## Development\n\nPull requests and contributors are welcome.\n\nspaczz is linted with [Flake8](https://flake8.pycqa.org/en/latest/), formatted with [Black](https://black.readthedocs.io/en/stable/), type-checked with [MyPy](http://mypy-lang.org/) (although this could benefit from improved specificity), tested with [Pytest](https://docs.pytest.org/en/stable/), automated with [Nox](https://nox.thea.codes/en/stable/), and built/packaged with [Poetry](https://python-poetry.org/). There are a few other development tools detailed in the noxfile.py, along with Git pre-commit hooks.\n\nTo contribute to spaczz\'s development, fork the repository then install spaczz and it\'s dev dependencies with Poetry. If you\'re interested in being a regular contributor please contact me directly.\n\n\n```python\npoetry install # Within spaczz\'s root directory.\n```\n\nI keep Nox and pre-commit outside of my poetry environment as part of my Python toolchain environments. With pre-commit installed you may also need to run the below to commit changes.\n\n\n```python\npre-commit install\n```\n\nThe only other package that will not be installed via Poetry but is used for testing and in-documentation examples is the spaCy medium English model (`en-core-web-md`). This will need to be installed separately. The command below should do the trick:\n\n\n```python\npoetry run python -m spacy download "en_core_web_md"\n```\n\n## References\n\n- Spaczz tries to stay as close to [spaCy](https://spacy.io/)\'s API as possible. Whenever it made sense to use existing spaCy code within spaczz this was done.\n- Fuzzy matching is performed using [RapidFuzz](https://github.com/maxbachmann/rapidfuzz).\n- Regexes are performed using the [regex](https://pypi.org/project/regex/) library.\n- The search algorithm for phrased-based fuzzy and similarity matching was heavily influnced by Stack Overflow user Ulf Aslak\'s answer in this [thread](https://stackoverflow.com/questions/36013295/find-best-substring-match).\n- Spaczz\'s predefined regex patterns were borrowed from the [commonregex](https://github.com/madisonmay/CommonRegex) package.\n- Spaczz\'s development and CI/CD patterns were inspired by Claudio Jolowicz\'s [*Hypermodern Python*](https://cjolowicz.github.io/posts/hypermodern-python-01-setup/) article series.\n',
    'author': 'Grant Andersen',
    'author_email': 'gandersen.codes@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gandersen101/spaczz',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Module for RegexPredef class."""
from __future__ import annotations

from typing import Any

import regex

from ._commonregex import _commonregex
from ..exceptions import RegexParseError


class RegexConfig:
    """Class for parsing regex patterns and housing predefined regex patterns.

    Will eventually includes methods for adding/removing predefined user patterns.

    Attributes:
        _predefs: Regex patterns available.
    """

    def __init__(self: RegexConfig, empty: bool = False) -> None:
        """Initializes the regex config.

        Args:
            empty: Whether to initialize the instance without predefined
                patterns or not. Will be more useful later once API
                is extended. Default is False.
        """
        if not empty:
            self._predefs = _commonregex
        else:
            self._predefs = {}

    def parse_regex(
        self: RegexConfig,
        regex_str: str,
        predef: bool = False,
    ) -> Any:
        """Parses a string into a regex pattern.

        Will treat string as a key name for a predefined regex
        if predef is True.

        Args:
            regex_str: String to compile into a regex pattern.
            predef: Whether regex should be interpreted as a key to
                a predefined regex pattern or not. Default is False.
                The included regexes are:
                "dates"
                "times"
                "phones"
                "phones_with_exts"
                "links"
                "emails"
                "ips"
                "ipv6s"
                "prices"
                "hex_colors"
                "credit_cards"
                "btc_addresses"
                "street_addresses"
                "zip_codes"
                "po_boxes"
                "ssn_number".

        Returns:
            A compiled regex pattern.

        Raises:
            RegexParseError: If regex compilation produces any errors.

        Example:
            >>> import regex
            >>> from spaczz.regex import RegexConfig
            >>> predef = RegexConfig()
            >>> pattern = predef.parse_regex("Test")
            >>> isinstance(pattern, type(regex.compile("type")))
            True
        """
        if predef:
            compiled_regex = self.get_predef(regex_str)
        else:
            try:
                compiled_regex = regex.compile(
                    regex_str,
                )
            except (regex._regex_core.error, TypeError, ValueError) as e:
                raise RegexParseError(e)
        return compiled_regex

    def get_predef(self: RegexConfig, predef: str) -> Any:
        """Returns a regex pattern from the predefined patterns available.

        Args:
            predef: The key name of a predefined regex pattern.

        Returns:
            A compiled regex pattern.

        Raises:
            ValueError: If the key does not exist in the predefined regex patterns.

        Example:
            >>> import regex
            >>> from spaczz.regex import RegexConfig
            >>> config = RegexConfig()
            >>> pattern = config.get_predef("phones")
            >>> isinstance(pattern, type(regex.compile("type")))
            True
        """
        predef_regex = self._predefs.get(predef)
        if predef_regex:
            return predef_regex
        else:
            raise ValueError(
                f"{predef} is not a regex pattern defined in this RegexConfig instance."
            )



"""Module for various utility functions."""
from collections import defaultdict
from functools import partial
from itertools import repeat, tee
from pathlib import Path
from typing import Any, DefaultDict, Iterable, Union


def ensure_path(path: Union[str, Path]) -> Path:
    """Ensure string is converted to a Path.

    Args:
        path: Anything. If string, it's converted to Path.

    Returns:
        Path or original argument.
    """
    if isinstance(path, str):
        return Path(path)
    else:
        return path


def nest_defaultdict(default_factory: Any, depth: int = 1) -> DefaultDict[Any, Any]:
    """Returns a nested `defaultdict`."""
    result = partial(defaultdict, default_factory)
    for _ in repeat(None, depth - 1):
        result = partial(defaultdict, result)
    return result()


def n_wise(iterable: Iterable[Any], n: int) -> Iterable[Any]:
    """Iterates over an iterables in slices of length n by one step at a time."""
    iterables = tee(iterable, n)
    for i in range(len(iterables)):
        for _ in range(i):
            next(iterables[i], None)
    return zip(*iterables)


def read_from_disk(path: Union[str, Path], readers: Any, exclude: Any) -> Path:
    """Reads a pipeline component from disk."""
    path = ensure_path(path)
    for key, reader in readers.items():
        # Split to support file names like meta.json
        if key.split(".")[0] not in exclude:
            reader(path / key)
    return path


def write_to_disk(path: Union[str, Path], writers: Any, exclude: Any) -> Path:
    """Writes a pipeline component to disk."""
    path = ensure_path(path)
    if not path.exists():
        path.mkdir()
    for key, writer in writers.items():
        # Split to support file names like meta.json
        if key.split(".")[0] not in exclude:
            writer(path / key)
    return path
